package cologne.eck.calendar_pea.util;

/*
 * PeaFactory - Production of Password Encryption Archives
 * Copyright (C) 2015  Axel von dem Bruch
 * 
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, 
 * or (at your option) any later version.
 * This library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * See:  http://www.gnu.org/licenses/gpl-2.0.html
 * You should have received a copy of the GNU General Public License 
 * along with this library.
 */

import static org.junit.jupiter.api.Assertions.*;

import java.io.File;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import cologne.eck.calendar_pea.data.PCalendar;
import cologne.eck.calendar_pea.data.PEvent;
import cologne.eck.calendar_pea.data.RecurringPEvent;
import cologne.eck.tools.ReadResources;
import cologne.eck.tools.TestLog;


class ReadICalTest  {
	
	private PCalendar[] cals;
	
	private static final String[] cal1DateStartStrings = {
			"2018-12-29T07:00",
			"2018-12-29T10:05",
			"2018-12-29T16:00",
			"2018-12-30T00:00",
			"2018-12-30T00:00",
			"2018-12-30T09:00",
			"2018-12-30T10:00",
			"2018-12-30T11:00",
			"2018-12-30T12:00",
			"2018-12-30T13:00",
			"2018-12-31T00:00",
			"2018-12-31T09:00"		
	};
	private static final String[] cal1DateEndStrings = {
			"2018-12-29T09:00", 
			"2018-12-29T12:05", 
			"2018-12-30T09:00", 
			null, 
			null, 
			"2018-12-30T10:00", 
			"2018-12-30T11:00", 
			"2018-12-30T12:00", 
			"2018-12-30T13:00", 
			"2018-12-30T14:00", 
			"2019-01-02T00:00", 
			"2019-01-01T10:00"
	};
	private static final String[] rDateStrings = {"2019-04-11", "2020-03-03"};
	private static final String[] exDateStrings = {"2019-04-01", "2019-04-28"};
	private static final String[][] byDayStrings = { {"+2MO", "-1SU"}, {"+1MO", "-1SU"},{"SU", "WE"}};
	

	@BeforeEach
	public void setUp() {
		byte[] calBytes = ReadResources.readExternFile(
				System.getProperty("user.dir") + File.separator + 
				"resources" + File.separator + "test" + File.separator + "test-calendar1.ics");
		
		cals = new ReadICal().readCalendarsFromBytes(calBytes, System.getProperty("user.dir") + File.separator + 
				"resources" + File.separator + "test" + File.separator + "test-calendar1.ics", true);
		if (cals == null) {
			TestLog.e(getClass(), "No calendars");
		}
	}

	@Test
	public void testReadCalendarsFromBytes() {
		
		assertNotNull(cals);
		
		PCalendar cal1 = cals[0];
		
		assertNotNull(cals[0]);
		assertNotNull(cals[0].getCalUid());
		assertNotNull(cals[0].getName());
		assertNotNull(cals[0].getEventList());
		
		ArrayList<PEvent> eventList = cal1.getEventList();
		for (int i = 0; i < eventList.size(); i++) {
			PEvent e = eventList.get(i);
			assertTrue(e.getSummary().startsWith("event" + (i+1)));
			assertEquals(e.getDateStart(), LocalDateTime.parse(cal1DateStartStrings[i]));		
			if (cal1DateEndStrings[i] != null) {
				assertEquals(e.getDateEnd(), LocalDateTime.parse(cal1DateEndStrings[i]));	
			}
			if (i < 4) {
				assertEquals("event description", e.getDescription());
			} else {
				assertNull(e.getDescription());
			}
			if ( i == 1 || i == 4 || i == 8 || i == 11) {
				assertEquals(e.getPriority(), 5);
			}
			if (i == 1 || i == 11) {
				assertEquals(e.getStatus(), "CONFIRMED");
			} else if (i == 4) {
				assertEquals(e.getStatus(), "CANCELLED");
			} else if (i == 8) {
				assertEquals(e.getStatus(), "TENTATIVE");
			}
			if (i == 3 || i== 4 || i == 10) {
				assertTrue(e.isAllDay());
			} else {
				assertFalse(e.isAllDay());
			}
			if (i == 2 || i == 10 || i == 11) {
				assertTrue(e.isSeveralDays());
			} else {
				assertFalse(e.isSeveralDays());
			}
			assertNotNull(e.getUid());
			assertNotNull(e.getTimeStamp());
			if (i > 3 && i < 10) {
				assertTrue(e.isRecurring());
				RecurringPEvent re = (RecurringPEvent) e;	
				
				if (i == 4 || i == 9) {
					assertEquals(re.getFrequency(), ChronoUnit.DAYS);
				} else if (i == 5 || i == 6) {
					assertEquals(re.getFrequency(), ChronoUnit.YEARS);
				} else if (i == 7) {
					assertEquals(re.getFrequency(), ChronoUnit.MONTHS);
				} else if (i == 8) {
					assertEquals(re.getFrequency(), ChronoUnit.WEEKS);
				}
				
				if (i != 5) {
					assertEquals(re.getInterval(), 3);
				} else {
					assertEquals(re.getInterval(), 1);
				}
				if (i == 7) {
					LocalDate[] rDates = re.getRecurrenceDates();
					for (int j=0; j<rDates.length;j++) {
						assertEquals(rDates[j], LocalDate.parse(rDateStrings[j]));
					}
					LocalDate[] exDates = re.getExceptionDates();
					for (int j=0; j<exDates.length;j++) {
						assertEquals(exDates[j], LocalDate.parse(exDateStrings[j]));
						//TestLog.o(XXXX.class, exDates[j]);
					}
				} else {
					assertNull(re.getExceptionDates());
					assertNull(re.getRecurrenceDates());
				}
				if ( i > 5 && i < 9) {
					String[] byDays = re.getByDay(); 
					for (int j = 0; j < byDays.length; j++) {
						if (i == 6) {
							assertEquals(re.getByDay()[j], byDayStrings[0][j]);
						} else if (i == 7 ) {
							assertEquals(re.getByDay()[j], byDayStrings[1][j]);
						} else if (i == 8 ) {
							assertEquals(re.getByDay()[j], byDayStrings[2][j]);
						} 
					}
				} else {
					assertNull(re.getByDay());
				}
				if (i == 6) {
					assertEquals(re.getByMonth()[0], 12);
				} else {
					assertNull(re.getByMonth());
				}
							
			} else {
				assertFalse(e.isRecurring());
			}
			//TestLog.o(XXXX.class, e.isRecurring() + " " + i);
		}		
		TestLog.o(ReadICalTest.class, "readCalendarsFromBytes()  ok");
	}

	@Test
	public void testGetNumberOfEvents() {
		assertNotNull(cals);
		assertEquals(cals[0].getEventList().size(), ReadICal.getNumberOfEvents());
		TestLog.o(ReadICalTest.class, "getNumberOfEvents()  ok");
	}
}
