package cologne.eck.image_pea;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.charset.Charset;
import java.util.Arrays;
import java.util.HashMap;

import cologne.eck.all_peas.data.PeaProperties;
import cologne.eck.tools.Converter;
import cologne.eck.tools.ReadResources;
import cologne.eck.tools.TestLog;


/*
 * EXIF data are stored in an image file directory (IFD), 
 * IFD0: contains some entries including index of SUB-IFD (Exif IFD), GPS-IFD and InteroperabilityIFD. 
 * Value are stored after the IFD if value is longer than 4 bytes. 
 * 
 * View EXIF data: http://exif.regex.info/exif.cgi
 * EXIF format: https://www.media.mit.edu/pia/Research/deepview/exif.html
 * 				https://www.cipa.jp/std/documents/e/DC-X008-Translation-2019-E.pdf
 * TIFF format: https://www.itu.int/itudoc/itu-t/com16/tiff-fx/docs/tiff6.pdf 
 * EXIF TAGs: https://exiftool.org/TagNames/EXIF.html
 */

class JPGMetadataHandler extends MetadataHandler {
	
	private StringBuilder sb = null;
	
	//String END_OF_STREAM = "FFD9";
	
	// Start of Stream FFD8 and EXIF Application marker FFE1
	private static final byte[] EXIF_START_MARKER = { (byte) 0xFF, (byte) 0xD8, (byte) 0xFF, (byte) 0xE1};
	//String EXIF = "457869660000";// Exit + 2 zero 0x45 78 69 66
	private static final byte[] EXIF = "Exif\0\0".getBytes();
	private static final byte[] B_ENDIAN = {0x4D, 0x4D};
	private static final byte[] L_ENDIAN = {0x49, 0x49};
	private static final byte[] TIFF_MARKER = {0x00, 0x2A};// 42
	
	private static boolean be = true;
	private static int exifSubIFDIndex = 0;
	private static int exifGPSIFDIndex = 0;
	private static int exifInteropIndex = 0;

	// Most privacy related metadata only
	private static final HashMap<Integer, String> IFD0 = new HashMap<Integer, String>();
	
	// GPS Attribute Information
	// index in array is tag (from 0 up to 1F /31
	private static final String[] GPS = {// 0
	"GPSVersionID",  "North or South", "Latitude", "East or West", "Longitude",
	"Altitude Ref.","Altitude","GPS time","GPS satellite","GPS receiver status", // 9
	"Measurement mode","Measurement precision","Speed unit","Speed of GPS receiver",
	"Ref. for direction of movement","Direction of movement","Ref. for direction of image",
	"Direction of image","Geodetic survey data","Ref. for latitude of dest.", // 19
	"Latitude of destination","Ref. for longitude of dest.","Longitude of destination",
	"Ref. for bearing of dest.","Bearing of destination","Ref. for distance to dest.",
	"Distance to destination","GPS processing method","Name of GPS area","GPS date",
	"GPS differential correction","Horizontal positioning error" };//31
	
	
/* FOR TESTING
	public static void main(String[] args) {
		//byte[] jpgBytes = ReadResources.readExternFile("/home/ax/TEST/Image/img.jpg");
		//String result = new JPGMetadataHandler().getJpgMetadata(jpgBytes);		
		String fileName = "/home/ax/TEST/Image/img9.jpg";
		String result = new JPGMetadataHandler().getMetadata(fileName);
		System.out.println(result);
	} */
	
	/**
	 * Fill the map with EXIF Tags and corresponding strings
	 */
	private void fillHashMap() {

		if (IFD0.isEmpty()) {
			// TIFF Rev. 6.0 Attribute Information Used in Exif		
			IFD0.put(0x0002,	"Interop. Version");
			IFD0.put(0x000b, "Processing Software");
			IFD0.put(0x0103, "Compression");
			IFD0.put(0x010d, "Document Name");
			IFD0.put(0x011d, "Page Name");
			IFD0.put(0x013c, "Host Computer");
			// D. Other tags
			IFD0.put(0x010e,	"Image Description");
			IFD0.put(0x010f, "Manufacturer of digicam");	
			IFD0.put(0x0110, "Model number of digicam");
			IFD0.put(0x0131,"Software version number"); 	
			IFD0.put(0x0132 , "Date/Time last modified"); 
			IFD0.put(0x8298, "Copyright");				
			IFD0.put(0x013B, "Artist");

			// Exif IFD Attribute Information (1) 		
			// D. Tags Relating to User Information
			IFD0.put(0x927C, "Manufacturer notes");
			IFD0.put(0x9286, "User comments");

			// E. Tag Relating to Related File Information
			IFD0.put(0xA004, "Related audio file");

			// F. Tags Relating to Date and Time
			IFD0.put(0x9003, "Date and time generation original");
			IFD0.put(0x9004, "Date and time generation digital");
			IFD0.put(0x9010, "Offset data of DateTime");
			IFD0.put(0x9011, "OffsetTimeOriginal");
			IFD0.put(0x9012, "OffsetTimeDigitized");
			IFD0.put(0x9286 , "User Comment");
			IFD0.put(0x9290, "DateTime subseconds");
			IFD0.put(0x9291, "subseconds original");
			IFD0.put(0x9292, "subseconds digitized");
			
		//	0x9c9b 	XPTitle 	int8u 	IFD0 	(tags 0x9c9b-0x9c9f are used by Windows Explorer; special characters in these values are converted to UTF-8 by default, or Windows Latin1 with the -L option. XPTitle is ignored by Windows Explorer if ImageDescription exists)
		//	0x9c9c 	XPComment 	int8u 	IFD0 	 
		//	0x9c9d 	XPAuthor 	int8u 	IFD0 	(ignored by Windows Explorer if Artist exists)
		//	0x9c9e 	XPKeywords 	int8u 	IFD0 	 
		//	0x9c9f 	XPSubject 	int8u 	IFD0 	 

			// G. Tags Relating to Picture-Taking Conditions
			IFD0.put(0xA300, "File source");
			IFD0.put(0xA40B, "Device settings description");

			// G2. Tags Relating to Shooting Situation
			IFD0.put(0x9400, "Temperature");
			IFD0.put(0x9401, "Humidity");
			IFD0.put(0x9402, "Pressure");
			IFD0.put(0x9403, "WaterDepth");
			IFD0.put(0x9404, "Acceleration");
			IFD0.put(0x9405, "CameraElevationAngle");

			// H. Other tags
			IFD0.put(0xA420, "Unique image ID");
			IFD0.put(0xA430, "Camera Owner Name");
			IFD0.put(0xA431, "Body Serial Number");
			IFD0.put(0xA432, "Lens Specification");
			IFD0.put(0xA433, "Lens Make");
			IFD0.put(0xA434, "Lens Model");
			IFD0.put(0xA435, "Lens Serial Number");		

			IFD0.put(0x8825,	"GPS offset");  // 34853 (8825.H)  GPS IFD
			IFD0.put(0xA005,	"Interoperability offset");  // 40965 (A005.H)  Interoperability IFD		
			IFD0.put(0x9000,	"Exif version");// 36864 (9000.H) EXIF version unknown type, Default = "0232”	
			
			IFD0.put(0xfde8, "OwnerName");	//  generated by Photoshop Camera RAW
			IFD0.put(0xfde9, "SerialNumber");	//  generated by Photoshop Camera RAW
			
			// Not shown notes:			
			IFD0.put(0x83bb, "IPTC-NAA metadata (not shown)");  //  read IPCT IFD  TODO
			IFD0.put(0x8568, "AFCP_IPTC metadata (not shown)");  //  read IPCT IFD  TODO
			IFD0.put(0x02bc, "XMP Application Notes (not shown)");  //  read IPCT IFD  TODO
			IFD0.put(0x927c, "Company metadata (not shown)");  //  read extra IFD of maker TODO
		}
	}
	
	
	/**
	 * Read an IFD as described in TIFF specification, 
	 * append data to StringBuilder and return the pointer 
	 * to the next IFD
	 * 
	 * @param fileBytes	the file as byte array
	 * @param offset	the index of the IFD to read
	 * @param ifdType	0: IFD0 and Sub-IFD, 1: GPS, 2: Interoperability
	 * @param inJPG		true: read from JPG file
	 * 
	 * @return	the index of the next IFD
	 */
	private int readIFD(byte[] fileBytes, int offset, int ifdType, boolean inJPG) {
		
		int extraVal = 12;
		if (inJPG == false) {
			extraVal = 0;
		}
		//System.out.println("offset: "+offset);
		int numberOfEntries = 0;
		if (be == true) {
			numberOfEntries = getBE(fileBytes[offset], fileBytes[offset + 1]);
		} else {
			numberOfEntries = getLE(fileBytes[offset], fileBytes[offset + 1]);
		}
		//System.out.println("Number of entries: " + numberOfEntries );
		
		int indexOfEntry = offset + 2; // for number of entries = 2 bytes
		for (int i = 0; i < numberOfEntries; i++) {
			byte[] entry = new byte[12];
			System.arraycopy(fileBytes, indexOfEntry, entry, 0, entry.length);
			//System.out.println("Entry " + i + ": " + Converter.bytes2hex(entry));			
			
			// Get the tag:
			String tagString = null;
			int tag = 0;
			if (be == true) {
				tag = getBE(entry[0], entry[1]);
			} else {
				tag = getLE(entry[0], entry[1]);
			}
			if (ifdType == 0) { // EXIF IFD
				if (tag == 0x8769) { // get offset of EXIF Sub-IFD
					byte[] longValue = new byte[4];
					System.arraycopy(entry, 8, longValue, 0, longValue.length);
					if (be == true) {
						exifSubIFDIndex =  Converter.bytes2intsBE(longValue)[0];
					} else {
						exifSubIFDIndex =  Converter.bytes2intsLE(longValue)[0];
					}
					//System.out.println("EXIF sub IFD: " + exifSubIFDIndex);
					indexOfEntry += 12;
					continue;
				} else if ( (tag == 0x014a || tag == 0xc6f5) && exifSubIFDIndex == 0) { // other exif sub IFD
					// read only when there is no other subIFD
					// Sony DSLR-A100 0x014a or ProfileIFD 0xc6f5
					byte[] longValue = new byte[4];
					System.arraycopy(entry, 8, longValue, 0, longValue.length);
					if (be == true) {
						exifSubIFDIndex =  Converter.bytes2intsBE(longValue)[0];
					} else {
						exifSubIFDIndex =  Converter.bytes2intsLE(longValue)[0];
					}
					//System.out.println("EXIF sub IFD: " + exifSubIFDIndex);
					indexOfEntry += 12;
					continue;
				} else if (tag == 0x8825) { // get offset of EXIF GEO-IFD
					byte[] longValue = new byte[4];
					System.arraycopy(entry, 8, longValue, 0, longValue.length);
					if (be == true) {
						exifGPSIFDIndex =  Converter.bytes2intsBE(longValue)[0];
					} else {
						exifGPSIFDIndex =  Converter.bytes2intsLE(longValue)[0];
					}
					//System.out.println("EXIF GPS: " + exifSubIFDIndex);
					indexOfEntry += 12;
					continue;
				} else if (tag == 0xA005) { // get offset of EXIF Interop-IFD
					byte[] longValue = new byte[4];
					System.arraycopy(entry, 8, longValue, 0, longValue.length);
					if (be == true) {
						exifInteropIndex =  Converter.bytes2intsBE(longValue)[0];
					} else {
						exifInteropIndex =  Converter.bytes2intsLE(longValue)[0];
					}
					//System.out.println("EXIF GPS: " + exifSubIFDIndex);
					indexOfEntry += 12;
					continue;
				} 
				tagString = IFD0.get(tag);
			} else if (ifdType == 1){ // GPS data
				if (tag < GPS.length) {
					tagString = GPS[tag];
				}
			} else if (ifdType == 2) { // Interoperability Identification
				if (tag == 1) {
					tagString = "Interoperability Identification";
				}
			}
			// TODO: process IPTC and XMP
			// 0x83bb 	IPTC-NAA 	int32u! 	IFD0 	--> IPTC Tags
			// 0x8568 	AFCP_IPTC 	- 	- 	--> IPTC Tags			
			// 0x02bc 	ApplicationNotes 	int8u! 	IFD0 	--> XMP Tags

			// Format:
			int typeShort = 0;
			if (be == true) {
				typeShort = getBE(entry[2], entry[3]);
			} else {
				typeShort = getLE(entry[2], entry[3]);
			}
//			System.out.println("Format: " + typeShort + " - as hex: " + Integer.toHexString(typeShort));
				
			// Number of values of type (not always byte):
			byte[] numBytes = new byte[4];
			System.arraycopy(entry, 4, numBytes, 0, numBytes.length);
			int numb = 0;
			if (be == true) {
				numb = Converter.bytes2intsBE(numBytes)[0];
			} else {
				numb = Converter.bytes2intsLE(numBytes)[0];
			}
//			System.out.println("Data length: " + numb);			
			
			String val = "";
			int offsetValue = 0;
			int numberOfBytes = 0;
			
			// Create the value depending on type and length
			switch(typeShort) {
			case 1: 
				//System.out.println("BYTE- 1 Byte");// unsigned integer
				numberOfBytes = 1;
				if (numberOfBytes * numb <= 4) {
					//System.out.println("VALUE" );
					for (int j = 0; j < numb; j++) {
						val += Byte.toUnsignedInt(entry[8 + j]) + " ";
					}		
				} else {
					// Offset:
					byte[] offsetOrValue = new byte[4];
					System.arraycopy(entry, 8, offsetOrValue, 0, offsetOrValue.length);
					if (be == true) {
						offsetValue = Converter.bytes2intsBE(offsetOrValue)[0];
					} else {
						offsetValue = Converter.bytes2intsLE(offsetOrValue)[0];
					}
					//System.out.println("Offset: " + offsetValue + ", len: " + numb);		
					for (int j = 0; j < numb; j++) {
						if ((offsetValue + extraVal + j) < fileBytes.length) {
							val += Byte.toUnsignedInt(fileBytes[offsetValue + extraVal + j]) + " ";
						} else {
							TestLog.ve(JPGMetadataHandler.class, "Invalid offset" + (offsetValue + extraVal + j), 5);
							return 0;
						}
					}				
				}
				break;
			case 2:
				//System.out.println("ASCII- 1 Byte");//  7-bit ASCII code; the last byte must be NUL (binary zero)
				numberOfBytes = 1;				
				if (numberOfBytes * numb <= 4) {
					byte[] byteVal = new byte[4];
					System.arraycopy(entry, 8, byteVal, 0, 4);
					val = new String(byteVal);
				} else {
					// Offset:
					byte[] offsetOrValue = new byte[4];
					System.arraycopy(entry, 8, offsetOrValue, 0, offsetOrValue.length);
					if (be == true) {
						offsetValue = Converter.bytes2intsBE(offsetOrValue)[0];
					} else {
						offsetValue = Converter.bytes2intsLE(offsetOrValue)[0];
					}	
					byte[] valBytes = new byte[numb];
					System.arraycopy(fileBytes, offsetValue + extraVal, valBytes, 0, valBytes.length);
					val = new String(valBytes);
				}
				// Always display metadata of this type,even if type is unknown: 
				if (tagString == null) {
					tagString = "UNKNOWN CONTENT " +  Integer.toHexString(tag);
				}
				break;
			case 3: // TODO unsigned
				//System.out.println("SHORT - 2 Bytes");// unsigned integer
				numberOfBytes = 2;
				if (numberOfBytes * numb <= 4) {
					//System.out.println("VALUE" );
					if (be == true) {
						val += getBE(entry[8], entry[8 + 1]) + " ";
					} else {
						val += getLE(entry[8], entry[8 + 1]) + " ";
					}
				} else {
					// Offset:
					byte[] offsetOrValue = new byte[4];
					System.arraycopy(entry, 8, offsetOrValue, 0, offsetOrValue.length);
					if (be == true) {
						offsetValue = Converter.bytes2intsBE(offsetOrValue)[0];
					} else {
						offsetValue = Converter.bytes2intsLE(offsetOrValue)[0];
					}
					//System.out.println("Offset: " + offsetValue + ", len: " + numb);	
					for (int j = 0; j < numb; j+=2) {
						if (be == true) {
							val += getBE(fileBytes[offsetValue + extraVal + j], fileBytes[offsetValue + extraVal + j + 1]) + " ";
						} else {
							val += getLE(fileBytes[offsetValue + extraVal + j], fileBytes[offsetValue + extraVal + j + 1]) + " ";
						}
					}		
				}
				break;
			case 4: // TODO unsigned
				//System.out.println("LONG - 4 Bytes");// unsigned integer
				numberOfBytes = 4;
				if (numberOfBytes * numb <= 4) {
					//System.out.println("VALUE" );
					byte[] longValue = new byte[4];
					System.arraycopy(entry, 8, longValue, 0, longValue.length);
					if (be == true) {
						val = "" + Converter.bytes2intsBE(longValue)[0];
					} else {
						val = "" + Converter.bytes2intsLE(longValue)[0];
					}
				} else {
					// Offset:
					byte[] offsetOrValue = new byte[4];
					System.arraycopy(entry, 8, offsetOrValue, 0, offsetOrValue.length);
					if (be == true) {
						offsetValue = Converter.bytes2intsBE(offsetOrValue)[0];
					} else {
						offsetValue = Converter.bytes2intsLE(offsetOrValue)[0];
					}
					//System.out.println("Offset: " + offsetValue + ", len: " + numb);	
					for (int j = 0; j < numb; j+=4) {
						if (be == true) {
							val += getBE(fileBytes[offsetValue + extraVal + j], fileBytes[offsetValue + extraVal + j + 1],
									fileBytes[offsetValue + extraVal + j + 2], fileBytes[offsetValue + extraVal + j + 3]) + " ";
						} else {
							val += getLE(fileBytes[offsetValue + extraVal + j], fileBytes[offsetValue + extraVal + j + 1],
									fileBytes[offsetValue + extraVal + j + 2], fileBytes[offsetValue + extraVal + j + 3]) + " ";
						}
					}
				}
				break;
			case 5: // TODO unsigned
				//System.out.println("RATIONAL - 8 Bytes");// the first represents the numerator of a
				// fraction; the second, the denominator Bruch Zähler-Nenner
				numberOfBytes = 8;
				// Offset:
				byte[] offsetOrValue = new byte[4];
				System.arraycopy(entry, 8, offsetOrValue, 0, offsetOrValue.length);
				if (be == true) {
					offsetValue = Converter.bytes2intsBE(offsetOrValue)[0];
				} else {
					offsetValue = Converter.bytes2intsLE(offsetOrValue)[0];
				}
				if (numb > 1) {
					if (numb <= 8) {
						if (offsetValue + numb < fileBytes.length) {
							for (int j = 0; j < numb; j+=4) {
								if (be == true) {
									val += getBE(fileBytes[offsetValue + extraVal + j], fileBytes[offsetValue + extraVal + j + 1],
											fileBytes[offsetValue + extraVal + j + 2], fileBytes[offsetValue + extraVal + j + 3]) + " ";
								} else {
									val += getLE(fileBytes[offsetValue + extraVal + j], fileBytes[offsetValue + extraVal + j + 1],
											fileBytes[offsetValue + extraVal + j + 2], fileBytes[offsetValue + extraVal + j + 3]) + " ";
								}
							}
						}
					} else {
						TestLog.ve(JPGMetadataHandler.class, "Could not get value for " + tag + ", length: " + numb, 5);
						val = "---Could not read---";
					}
				} else {
					int numerator = 0;
					int  denominator = 0;
					if (be == true) {
						numerator = getBE(fileBytes[offsetValue + extraVal], fileBytes[offsetValue + extraVal + 1],
								fileBytes[offsetValue + extraVal + 2], fileBytes[offsetValue + extraVal + 3]);
						denominator = getBE(fileBytes[offsetValue + extraVal + 4], fileBytes[offsetValue + extraVal + 1 + 4],
								fileBytes[offsetValue + extraVal + 2 + 4], fileBytes[offsetValue + extraVal + 3 + 4]);
					} else {
						numerator = getLE(fileBytes[offsetValue + extraVal], fileBytes[offsetValue + extraVal + 1],
								fileBytes[offsetValue + extraVal + 2], fileBytes[offsetValue + extraVal + 3]);
						denominator = getLE(fileBytes[offsetValue + extraVal + 4], fileBytes[offsetValue + extraVal + 1 + 4],
								fileBytes[offsetValue + extraVal + 2 + 4], fileBytes[offsetValue + extraVal + 3 + 4]);
					}
					val = numerator + " / " + denominator;
				}
				break;
			case 6: 
				//System.out.println("SBYTE- 1 Byte");// signed integer
				numberOfBytes = 1;
				if (numberOfBytes * numb <= 4) {
					//System.out.println("VALUE" );
					for (int j = 0; j < numb; j++) {
						val += entry[8 + j] + " ";
					}		
				} else {
					if (numb < 256) {
						// Offset:
						byte[] offsetBytes = new byte[4];
						System.arraycopy(entry, 8, offsetBytes, 0, offsetBytes.length);
						if (be == true) {
							offsetValue = Converter.bytes2intsBE(offsetBytes)[0];
						} else {
							offsetValue = Converter.bytes2intsLE(offsetBytes)[0];
						}
						byte[] testArray = new byte[numb];
						System.arraycopy(fileBytes, offsetValue + extraVal, testArray, 0, testArray.length);
						if (isASCII(new String(testArray)) == true) {
							val = new String(testArray, Charset.forName("US-ASCII"));
						} else {
						//System.out.println("Offset: " + offsetValue + ", len: " + numb);		
							for (int j = 0; j < numb; j++) {
								val += fileBytes[offsetValue + extraVal + j] + " ";
							}				
						}
					} else {
						TestLog.ve(JPGMetadataHandler.class, tagString + " - Long array of type unsigned byte is not printed, length: " + numb);
						val = "---Could not read---";
					}
				}
				break;
			case 7: 
				//System.out.println("UNDEFINED- 1 Byte");//  7-bit ASCII code; the last byte must be NUL (binary zero)
				numberOfBytes = 1;
				if (numberOfBytes * numb <= 4) {
					//System.out.println("VALUE" );
					for (int j = 0; j < numb; j++) {
						val += Byte.toUnsignedInt(entry[8 + j]) + " ";
					}		
				} else {
					if (numb < 64) { // print as bytes
						// Offset:
						offsetOrValue = new byte[4];
						System.arraycopy(entry, 8, offsetOrValue, 0, offsetOrValue.length);
						if (be == true) {
							offsetValue = Converter.bytes2intsBE(offsetOrValue)[0];
						} else {
							offsetValue = Converter.bytes2intsLE(offsetOrValue)[0];
						}
						//System.out.println("Offset: " + offsetValue + ", len: " + numb);	
						byte[] tryString = new byte[numb];
						System.arraycopy(fileBytes, offsetValue + extraVal, tryString, 0, tryString.length);
						if (isASCII(new String(tryString)) == true) {
							val = new String(tryString, Charset.forName("US-ASCII"));
						} else {
							for (int j = 0; j < numb; j++) {
								val += Byte.toUnsignedInt(fileBytes[offsetValue + extraVal + j]) + " ";
							}				
						}
					} else if (numb < 256) { // try to print as UTF-8
						
						byte[] tryString = new byte[numb];
						System.arraycopy(fileBytes, offsetValue + extraVal, tryString, 0, tryString.length);
						if (isASCII(new String(tryString)) == true) {
							val = new String(tryString, Charset.forName("US-ASCII"));
						} else {
							val += " - " + new String(tryString, PeaProperties.getCharset());
						}
					} else {
						TestLog.ve(JPGMetadataHandler.class, tagString + " - Long array of unknown type is not printed, length: " + numb);
						val = "---Could not read---";
					}
				}
				break;
			case 8: 
				//System.out.println("SSHORT - 2 Bytes");// signed integer
				if (numberOfBytes * numb <= 4) {
					//System.out.println("VALUE" );
					if (be == true) {
						val += getBE(entry[8], entry[8 + 1]) + " ";
					} else {
						val += getLE(entry[8], entry[8 + 1]) + " ";
					}
				} else {
					// Offset:
					byte[] offsetShort = new byte[4];
					System.arraycopy(entry, 8, offsetShort, 0, offsetShort.length);
					if (be == true) {
						offsetValue = Converter.bytes2intsBE(offsetShort)[0];
					} else {
						offsetValue = Converter.bytes2intsLE(offsetShort)[0];
					}
					//System.out.println("Offset: " + offsetValue + ", len: " + numb);	
					for (int j = 0; j < numb; j+=2) {
						if (be == true) {
							val += getBE(fileBytes[offsetValue + extraVal + j], fileBytes[offsetValue + extraVal + j + 1]) + " ";
						} else {
							val += getLE(fileBytes[offsetValue + extraVal + j], fileBytes[offsetValue + extraVal + j + 1]) + " ";
						}
					}		
				}
				numberOfBytes = 2;
				break;
			case 9: 
				//System.out.println("SLONG - 4 Bytes");// signed integer
				if (numberOfBytes * numb <= 4) {
					//System.out.println("VALUE" );
					byte[] longValue = new byte[4];
					System.arraycopy(entry, 8, longValue, 0, longValue.length);
					if (be == true) {
						val = "" + Converter.bytes2intsBE(longValue)[0];
					} else {
						val = "" + Converter.bytes2intsLE(longValue)[0];
					}
				} else {
					// Offset:
					byte[] offsetOoInts = new byte[4];
					System.arraycopy(entry, 8, offsetOoInts, 0, offsetOoInts.length);
					if (be == true) {
						offsetValue = Converter.bytes2intsBE(offsetOoInts)[0];
					} else {
						offsetValue = Converter.bytes2intsLE(offsetOoInts)[0];
					}
					//System.out.println("Offset: " + offsetValue + ", len: " + numb);	
					for (int j = 0; j < numb; j+=4) {
						if (be == true) {
							val += getBE(fileBytes[offsetValue + extraVal + j], fileBytes[offsetValue + extraVal + j + 1],
									fileBytes[offsetValue + extraVal + j + 2], fileBytes[offsetValue + extraVal + j + 3]) + " ";
						} else {
							val += getLE(fileBytes[offsetValue + extraVal + j], fileBytes[offsetValue + extraVal + j + 1],
									fileBytes[offsetValue + extraVal + j + 2], fileBytes[offsetValue + extraVal + j + 3]) + " ";
						}
					}
				}
				numberOfBytes = 4;
				break;
			case 10: 
				//System.out.println("SRATIONAL - 8 Bytes");// the first represents the numerator of a		
				numberOfBytes = 8;// fraction; the second, the denominator Bruch Zähler-Nenner
				// Offset:
				byte[] offsetOfULong = new byte[4];
				System.arraycopy(entry, 8, offsetOfULong, 0, offsetOfULong.length);
				if (be == true) {
					offsetValue = Converter.bytes2intsBE(offsetOfULong)[0];
				} else {
					offsetValue = Converter.bytes2intsLE(offsetOfULong)[0];
				}
				if (numb > 1) {
					TestLog.ve(JPGMetadataHandler.class, "Could not get value for " + tag, 5);
				} else {
					int numerator = 0;
					int  denominator = 0;
					if (be == true) {
						numerator = getBE(fileBytes[offsetValue + extraVal], fileBytes[offsetValue + extraVal + 1],
								fileBytes[offsetValue + extraVal + 2], fileBytes[offsetValue + extraVal + 3]);
						denominator = getBE(fileBytes[offsetValue + extraVal + 4], fileBytes[offsetValue + extraVal + 1 + 4],
								fileBytes[offsetValue + extraVal + 2 + 4], fileBytes[offsetValue + extraVal + 3 + 4]);
					} else {
						numerator = getLE(fileBytes[offsetValue + extraVal], fileBytes[offsetValue + extraVal + 1],
								fileBytes[offsetValue + extraVal + 2], fileBytes[offsetValue + extraVal + 3]);
						denominator = getLE(fileBytes[offsetValue + extraVal + 4], fileBytes[offsetValue + extraVal + 1 + 4],
								fileBytes[offsetValue + extraVal + 2 + 4], fileBytes[offsetValue + extraVal + 3 + 4]);
					}
					val = numerator + " / " + denominator;
				}
				break;
			case 11: 
				//System.out.println("FLOAT - 4 Bytes");//  IEEE format
				numberOfBytes = 4;
				if (numberOfBytes * numb <= 4) {
					//System.out.println("VALUE" );
					byte[] longValue = new byte[4];
					System.arraycopy(entry, 8, longValue, 0, longValue.length);
					int intVal = 0;
					if (be == true) {
						intVal = Converter.bytes2intsBE(longValue)[0];
					} else {
						intVal = Converter.bytes2intsLE(longValue)[0];
					}
					val = "" + Float.intBitsToFloat(intVal);
				}
				break;
			case 12: 
				//System.out.println("DOUBLE - 8 Bytes");//  IEEE format	
				numberOfBytes = 8;
				// Offset:
				byte[] offsetOfDouble = new byte[4];
				System.arraycopy(entry, 8, offsetOfDouble, 0, offsetOfDouble.length);
				if (be == true) {
					offsetValue = Converter.bytes2intsBE(offsetOfDouble)[0];
				} else {
					offsetValue = Converter.bytes2intsLE(offsetOfDouble)[0];
				}
				byte[] doubleBytes = new byte[numb*8];
				System.arraycopy(fileBytes, offsetValue + extraVal, doubleBytes, 0, doubleBytes.length);
				val = "" + ByteBuffer.wrap(doubleBytes).getDouble();
				break;
			default:
				//System.out.println("UNKNOWN TYPE: " + typeShort);
				numberOfBytes = 1;
				if (numberOfBytes * numb <= 4) {
					//System.out.println("VALUE" );
					for (int j = 0; j < numb; j++) {
						val += Byte.toUnsignedInt(entry[8 + j]) + " ";
					}		
				} else {
					if (numb < 64) {
						if (tagString == null) {
							tagString = "UNKNOWN TYPE";
						}
						// Offset:
						byte[] offsetUnknown = new byte[4];
						System.arraycopy(entry, 8, offsetUnknown, 0, offsetUnknown.length);
						if (be == true) {
							offsetValue = Converter.bytes2intsBE(offsetUnknown)[0];
						} else {
							offsetValue = Converter.bytes2intsLE(offsetUnknown)[0];
						}
						if (offsetValue + extraVal + numb < fileBytes.length) {
							byte[] tryString = new byte[numb];
							System.arraycopy(fileBytes, offsetValue + extraVal, tryString, 0, tryString.length);
							if (isASCII(new String(tryString)) == true) {
								val = new String(tryString, Charset.forName("US-ASCII"));
							} else {
								if (numb < 128) {
									//System.out.println("Offset: " + offsetValue + ", len: " + numb);		
									for (int j = 0; j < numb; j++) {
										val += Byte.toUnsignedInt(fileBytes[offsetValue + extraVal + j]) + " ";
									}				
								}
							}
						}
					} else {
						TestLog.e(JPGMetadataHandler.class, tagString + " - Long array of unknown type is not printed, length: " + numb);
					}
				}
				break;
			}			

			if (tagString == null) {
				//tagString =  Integer.toHexString(tag);
			} else {
				if (val == null || val.length() == 0) {
					val = "--Couldn't read data--";
				}
				sb.append(tagString + ": " + val +"\n");
			}					
			indexOfEntry += 12;		
		}

		int valueOffset = getBE(fileBytes[indexOfEntry], fileBytes[indexOfEntry + 1], fileBytes[indexOfEntry + 2], fileBytes[indexOfEntry + 3]);
		return valueOffset;
	}
	
	/**
	 * Get file metadata (size, name...) and EXIF data from a file
	 * 
	 * @param fileName	name of the file with absolute path
	 * 
	 * @return a multiline string of metadata
	 */
	@Override
	public String getMetadata(String fileName) {	
		
		byte[] jpgBytes = ReadResources.readExternFile(fileName);	
		return getMetadata(jpgBytes);
	}

	/**
	 * Get EXIF data from a file without file metadata such as size
	 * 
	 * @param imageBytes	content of file as byte array
	 * @param inJPG			true if in JPG or TIFF, false if in other format
	 * 						- that means, the start marker is not present, read from endian value
	 * 
	 * @return a multiline string of metadata
	 */
	protected String getMetadataAllFormat(byte[] imageBytes, boolean inJPG) {
		
		if (sb == null) {
			sb = new StringBuilder();
		}
		
		fillHashMap();		
		exifSubIFDIndex = 0;
		exifGPSIFDIndex = 0;
		exifInteropIndex = 0;
		
		if (sb == null) {
			sb = new StringBuilder();
		}
		int jpgHeaderLen = 0;
		
		if (inJPG == true) {
			// check if this is JPG: first bytes ffd8,last bytes ffd9
			byte[] startMarker = new byte[4];
			System.arraycopy(imageBytes, 0, startMarker, 0, startMarker.length);
			//System.out.println("Start Marker: " + Converter.bytes2hex(startMarker));
			if (Arrays.equals(startMarker, EXIF_START_MARKER)) {
				//			System.out.println("EXIF start");
			} else {
				TestLog.ve(JPGMetadataHandler.class, "Not exif data");
				sb.append("Not exif data\n");
				return new String(sb);
			}
			int size = getBE(imageBytes[startMarker.length], imageBytes[startMarker.length +1]);

			byte[] exif00 = new byte[EXIF.length];
			System.arraycopy(imageBytes,  startMarker.length + 2, exif00, 0, EXIF.length);
			if (Arrays.equals(EXIF,  exif00)) {
				//			System.out.println("Is exif");
			} else {
				TestLog.ve(JPGMetadataHandler.class, "Not exif: " + Converter.bytes2hex(exif00) + " - " + new String(exif00));
				sb.append("Not EXIF\n");
				return new String(sb);
			}
			//System.out.println("EXIF00: " + Converter.bytes2hex(exif00) + " - " + new String(exif00));
			byte[] exifData = new byte[size - 2 - EXIF.length]; // includes size: 2 bytes
			System.arraycopy(imageBytes, startMarker.length + EXIF.length + 2, exifData, 0, exifData.length);		

			jpgHeaderLen = startMarker.length + EXIF.length + 2;
		} else {
			// JPEG APP1 marker, length, and the "Exif ID code" described in 4.7.2(C), 
			// i.e., "Exif", NULL, and padding byte, are not included
			jpgHeaderLen = 0;
		}
		
		// Tiff file header 8 bytes: 2 endian, 2 fileNumber (42), 4 offset
		// check little or big endian:
		byte[] endian = new byte[2];
		endian[0] = imageBytes[jpgHeaderLen];
		endian[1] = imageBytes[jpgHeaderLen + 1];
		if (Arrays.equals(endian,  L_ENDIAN)) {
			TestLog.v(JPGMetadataHandler.class, "Little endian");
			be = false;
		} else if (Arrays.equals(endian,  B_ENDIAN)) {
			TestLog.v(JPGMetadataHandler.class, "Big endian");
			be = true;
		} else {
			TestLog.ve(JPGMetadataHandler.class, "Invald endian: " + Converter.bytes2hex(endian));
			return null;
		}
		
		byte[] fixedNumber = new byte[2];
		fixedNumber[0] = imageBytes[jpgHeaderLen + 2];
		fixedNumber[1] = imageBytes[jpgHeaderLen + 3];
		if (Arrays.equals(fixedNumber,  TIFF_MARKER)) {
			TestLog.v(JPGMetadataHandler.class, "TIFF file");
		}
		
		byte[] offset = new byte[4]; // offset of first IFD after the header, word boundary
		offset[0] = imageBytes[jpgHeaderLen + 4];
		offset[1] = imageBytes[jpgHeaderLen + 5];
		offset[2] = imageBytes[jpgHeaderLen + 6];
		offset[3] = imageBytes[jpgHeaderLen + 7];
		int offsetOfFirstIFD = 0;
		if (be == true) {
			offsetOfFirstIFD = Converter.bytes2intsBE(offset)[0];// + startMarker.length + EXIF.length + 10;
		} else {
			offsetOfFirstIFD = Converter.bytes2intsLE(offset)[0];// + startMarker.length + EXIF.length + 10;
		}
		//System.out.println("Offset: " +  offsetOfFirstIFD + ", + headerLen: " + (jpgHeaderLen + offsetOfFirstIFD));
		sb.append("EXIF data:" + "\n");
				
		readIFD(imageBytes, jpgHeaderLen + offsetOfFirstIFD, 0, inJPG);
		//System.out.println("NEXT OFFSET: " + nextIFDOffset);
		
		if (exifSubIFDIndex > 0) {
			sb.append("\nMore EXIF data:" + "\n");
			readIFD(imageBytes, exifSubIFDIndex + jpgHeaderLen, 0, inJPG);
		} else {
			sb.append("\nNo further EXIF data" + "\n");
		}
		if (exifGPSIFDIndex > 0) {
			sb.append("\nEXIF GPS data:" + "\n");
			readIFD(imageBytes, exifGPSIFDIndex + jpgHeaderLen, 1, inJPG);
		} else {
			sb.append("\nNo GPS data" + "\n");
		}
		if (exifInteropIndex > 0) {
			sb.append("\nEXIF Interoperability:" + "\n");
			readIFD(imageBytes, exifInteropIndex + jpgHeaderLen, 2, inJPG);
		}	
		return new String(sb);
	}
	/**
	 * Get short for big endian
	 * @param one	first (left) byte
	 * @param two	second byte
	 * @return	short for big endian
	 */
	private static int getBE(byte one, byte two) {
		return ((one & 0xff) << 8) | (two & 0xff);
	}
	/**
	 * Get short for little endian
	 * @param one	first (left) byte
	 * @param two	second byte
	 * @return	short for little endian
	 */
	private static int getLE(byte one, byte two) {
		return ((two & 0xff) << 8) | (one & 0xff);
	}
	/**
	 * Get int for big endian
	 * @param one	first (left) byte
	 * @param two	second byte
	 * @param three	third byte
	 * @param four	fourth byte
	 * @return	int for big endian
	 */
	private static int getBE(byte one, byte two, byte three, byte four) {
		byte[] byteArray = new byte[4];
		byteArray[0] = one;
		byteArray[1] = two;
		byteArray[2] = three;
		byteArray[3] = four;
		ByteBuffer bb = ByteBuffer.wrap(byteArray); // big-endian 
		int baInt = bb.getInt();
		return baInt;
	}
	/**
	 * Get int for little endian
	 * @param one	first (left) byte
	 * @param two	second byte
	 * @param three	third byte
	 * @param four	fourth byte
	 * @return	int for little endian
	 */
	private static int getLE(byte one, byte two, byte three, byte four) {
		byte[] byteArray = new byte[4];
		byteArray[0] = one;
		byteArray[1] = two;
		byteArray[2] = three;
		byteArray[3] = four;
		ByteBuffer bb = ByteBuffer.wrap(byteArray); // big-endian 
		bb.order( ByteOrder.LITTLE_ENDIAN);
		int baInt = bb.getInt();
		return baInt;
	}

	@Override
	public String getMetadata(byte[] imageBytes) {
		return getMetadataAllFormat(imageBytes, true);
	}
}
