package cologne.eck.image_pea;

import java.io.File;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.attribute.BasicFileAttributes;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.TimeZone;

import cologne.eck.tools.TestLog;

abstract class MetadataHandler {
	
	
	/**
	 * Get metadata from a file without file metadata such as size
	 * 
	 * @param imageBytes	content of file as byte array
	 * 
	 * @return a multiline string of metadata
	 */
	public abstract String getMetadata(byte[] imageBytes);
	
	/**
	 * Get file metadata (size, name...) and specific metadata from a file
	 * 
	 * @param fileName	name of the file with absolute path
	 * 
	 * @return a multiline string of metadata
	 */
	public abstract String getMetadata(String fileName);
	
	/**
	 * Get metadata of a file (size,last modified...)
	 * 
	 * @param fileName	name of the file
	 * 
	 * @return	metadata as String
	 */
	protected static String getFileMetadata(String fileName) {
		StringBuilder fileSb = new StringBuilder();
		File file = new File(fileName);
		if ( ! file.exists()) {
			TestLog.e(JPGMetadataHandler.class, "Does not exist: " + fileName);
			fileSb.append("File does not exist: " + fileName);
			return new String(fileSb) + "\nERROR";
		}
		if ( ! file.canRead()) {
			TestLog.e(JPGMetadataHandler.class, "Can't read: " + fileName);
			fileSb.append("Can not read file: " + fileName);
			return new String(fileSb) + "\nERROR";
		}
		if ( file.isDirectory()) {
			TestLog.e(JPGMetadataHandler.class, "Is directory: " + fileName);
			fileSb.append("File is directory: " + fileName + " - must be a file");
			return new String(fileSb) + "\nERROR";
		}
		
		Path path = file.toPath();
		String mimeType = "unknown";
		String creationTime = "couldn't read";
		String lastAccess = "couldn't read";		
	    try {
			mimeType = Files.probeContentType(path);
			BasicFileAttributes attr = Files.readAttributes(path, BasicFileAttributes.class);
			creationTime = LocalDateTime.ofInstant(attr.creationTime().toInstant(), ZoneId.systemDefault()).toString();
			lastAccess = LocalDateTime.ofInstant(attr.lastAccessTime().toInstant(), ZoneId.systemDefault()).toString();
		} catch (IOException e) {
			TestLog.ve(JPGMetadataHandler.class, e.getLocalizedMessage() + " - " + e.toString());
		}
		LocalDateTime ldt = 
		        LocalDateTime.ofInstant(Instant.ofEpochMilli(file.lastModified()), 
                        TimeZone.getDefault().toZoneId());  
		String timeString = ldt.format(DateTimeFormatter.ISO_LOCAL_DATE_TIME);

		fileSb.append( fileName + "\nSize: " + file.length() 
		+ "\nCreated: " + creationTime + "\nLast access: " + lastAccess
			+ "\nLast Modified: " + timeString + "\nMIME type: " + mimeType + "\n\n");
		return new String(fileSb);		
	}
	
	/**
	 * Check if a String can be encoded as ASCII
	 * 
	 * @param testString	the String to test
	 * 
	 * @return	true: is ASCII
	 */
	public boolean isASCII(String testString) {
		return Charset.forName("US-ASCII").newEncoder().canEncode(new String(testString));
	}

}
