package cologne.eck.peafactory.control;

/*
 * Peafactory - Production of Password Encryption Archives
 * Copyright (C) 2015  Axel von dem Bruch
 * 
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, 
 * or (at your option) any later version.
 * This library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * See:  http://www.gnu.org/licenses/gpl-2.0.html
 * You should have received a copy of the GNU General Public License 
 * along with this library.
 */

/**
 * Managing of entries of the jar file (Password Encrypted Archive).
 * The pea should contain only the required classes. 
 */

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.util.ArrayList;
import java.util.jar.*;
import java.util.regex.Matcher;

import org.bouncycastle.crypto.digests.Blake3Digest;
import org.bouncycastle.crypto.digests.SkeinDigest;
import org.bouncycastle.crypto.digests.WhirlpoolDigest;
import org.bouncycastle.crypto.engines.DSTU7624Engine;
import org.bouncycastle.crypto.engines.LEAEngine;
import org.bouncycastle.crypto.engines.RC6Engine;
import org.bouncycastle.crypto.engines.Shacal2Engine;
import org.bouncycastle.crypto.engines.TwofishEngine;

import cologne.eck.peafactory.control.data_type.CalendarType;
import cologne.eck.peafactory.control.data_type.DataType;
import cologne.eck.peafactory.control.data_type.EditorType;
import cologne.eck.peafactory.control.data_type.FileType;
import cologne.eck.peafactory.control.data_type.ImageType;
import cologne.eck.peafactory.crypto.*;
import cologne.eck.peafactory.crypto.kdf.*;
import cologne.eck.tools.ExceptionHelper;
import cologne.eck.tools.PropertyHandler;
import cologne.eck.tools.TestLog;
import cologne.eck.tools.UnexpectedValueException;
import cologne.eck.tools.WriteResources;


public class CreateJarFile {
	
	private static ArrayList<String> langsToAdd = new ArrayList<String>();
	
	//All JarEntry's names should NOT begin with "/".
	//Directory names must end with a slash '/'
	//All paths must use '/' style slashes, not '\'
	
	protected static void createJarFile() { 
		
		DataType dataType = DataType.getCurrentType();

		String currentJarFileName = JarStuff.getJarFileName();
		currentJarFileName = "peas" + File.separator + currentJarFileName;

		// replace with File.seperator to do some normal file operations
		if (currentJarFileName.contains("/") || currentJarFileName.contains("\\")) {
			currentJarFileName = currentJarFileName.replaceAll("/", File.separator);
			if (currentJarFileName.contains(Matcher.quoteReplacement("\\")) ){
				currentJarFileName = currentJarFileName.replaceAll( Matcher.quoteReplacement("\\"), File.separator  );
			}
			// if last directory not exist: create with all parent directories
			int index1 =  currentJarFileName.lastIndexOf( currentJarFileName.lastIndexOf(File.separator) -1 ) + 1;// if only one exists: return -1
			String lastDir = currentJarFileName.substring( index1 , currentJarFileName.lastIndexOf(File.separator));
			if ( ! new File(lastDir).exists()) {
				if (! new File(lastDir).mkdirs() ) {
					TestLog.e(CreateJarFile.class, "CreateJarFile.java: Can not create directories: " + lastDir);
					System.exit(3);
				}
			}
		}
		// operation on jar-file:		
		File jarFile = new File(currentJarFileName);
		// warn if file already exists
		if ( jarFile.exists() ) {
			TestLog.o(CreateJarFile.class, "Warning: existing jar-archive will be overwritten");
		}
		
		// Name of pea package
		String sep = File.separator;
		String pathToAllPeaClasses = "bin" + sep + "cologne" + sep + "eck" + sep 
				+ "all_peas" + sep;
		String pathToToolsClasses = "bin" + sep + "cologne" + sep + "eck" + sep 
				+ "tools" + sep;
		String pathToCryptoClasses = "bin" + sep + "cologne" + sep + "eck" + sep 
				+ "peafactory" + sep + "crypto" + sep;
		String pathToEckCryptoClasses = "bin" + sep + "cologne" + sep + "eck" + sep 
				+ "crypto" + sep;
		String pathToEckClasses = "bin" + sep + "cologne" + sep + "eck" + sep; 
		String pathToBcClasses = "bin" + sep + "org" + sep + "bouncycastle" + sep;

		// Manifest:
		Manifest manifest = new Manifest();
		Attributes att = manifest.getMainAttributes();		
		att.put(Attributes.Name.MANIFEST_VERSION, "1.0");		
		att.put(Attributes.Name.MAIN_CLASS, "start.Start");


		//-----------------
		// add the classes
		//
		ArrayList <File> classFiles = new ArrayList<File>();

		//==============================================
		// FIXED classes for all peas:
		//==============================================
		
		//===========================
		// peafactory classes:
		classFiles.add( new File("start/Start.class"));
		classFiles.add( new File("settings/PeaSettings.class"));
		
		classFiles.add( new File(pathToAllPeaClasses + "control/PeaControl.class"));	
		classFiles.add( new File(pathToAllPeaClasses + "control/PathFileManager.class"));
		classFiles.add( new File(pathToAllPeaClasses + "control/ContentUtilities.class"));
		classFiles.add( new File(pathToAllPeaClasses + "control/FileRecipient.class"));
		classFiles.add( new File(pathToAllPeaClasses + "control/CommandLineHandler.class"));
		classFiles.add( new File(pathToAllPeaClasses + "control/CommandLineHandler$EventDispatchThread.class"));
		classFiles.add( new File(pathToAllPeaClasses + "control/UpdateChecker.class"));
		classFiles.add( new File(pathToAllPeaClasses + "control/Translator.class"));
		
		classFiles.add( new File(pathToAllPeaClasses + "data/PeaProperties.class"));
		classFiles.add( new File(pathToAllPeaClasses + "data/Attachments.class"));
		classFiles.add( new File(pathToAllPeaClasses + "data/AttachmentHandler.class"));
		
		classFiles.add( new File(pathToAllPeaClasses + "bridge/VMBridge.class"));
		classFiles.add( new File(pathToAllPeaClasses + "vm_specific/JREBridge.class"));
		classFiles.add( new File(pathToAllPeaClasses + "vm_specific/JREProperties.class"));
		
		classFiles.add( new File(pathToAllPeaClasses + "files/ExecutionTimeObserver.class"));
		classFiles.add( new File(pathToAllPeaClasses + "files/FileComposer.class"));
		classFiles.add( new File(pathToAllPeaClasses + "files/FileModel.class"));
		classFiles.add( new File(pathToAllPeaClasses + "files/PeaCheckBox.class"));
		classFiles.add( new File(pathToAllPeaClasses + "files/PeaCheckBox$1.class"));// for remove Action
		classFiles.add( new File(pathToAllPeaClasses + "files/PeaDirCheckBox.class"));
		classFiles.add( new File(pathToAllPeaClasses + "files/FileTypePanel.class"));
		classFiles.add( new File(pathToAllPeaClasses + "files/FileTypePanel$ProgressTask.class")); // inner class
		classFiles.add( new File(pathToAllPeaClasses + "files/CloudFilePanel.class"));
		classFiles.add( new File(pathToAllPeaClasses + "files/FilePanelHolder.class"));
		classFiles.add( new File(pathToAllPeaClasses + "files/InvalidPanel.class"));
		
		classFiles.add( new File(pathToAllPeaClasses + "gui/PswDialogView.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/PswDialogView$1.class"));// MouseAdapter for eyeButton
		classFiles.add( new File(pathToAllPeaClasses + "gui/PeaLockFrame.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/PswDialogView$WhileTypingThread.class")); // inner class
		classFiles.add( new File(pathToAllPeaClasses + "gui/CharTable.class"));		
		classFiles.add( new File(pathToAllPeaClasses + "gui/KeyButton.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/CloseOnEscapeAction.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/ScrollbarAction.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/BorderButtonFocusListener.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/CurrentFileDisplay.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/menu/PathFileSetting.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/PeaFileChooser.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/menu/HelpMenu.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/menu/SettingMenu.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/menu/PeaMenuBar.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/PeaBorderFactory.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/NewPasswordDialog.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/Keyboard.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/MultiFileSelectionDialog.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/CursorManager.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/HtmlFileViewer.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/PeaDialog.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/PeaDialog$1.class"));// JPasswordDialog
		classFiles.add( new File(pathToAllPeaClasses + "gui/PeaDialog$2.class"));// ActionListener in link message
		classFiles.add( new File(pathToAllPeaClasses + "gui/IconManager.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/menu/ViewMenu.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/Toaster.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/Toaster$1.class"));// Thread
		classFiles.add( new File(pathToAllPeaClasses + "gui/KeyFileDialogs.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/CryptoSetter.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/kdf_setting/Argon2AdvancedPanel.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/kdf_setting/CatenaAdvancedPanel.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/kdf_setting/AllOtherAdvancedPanel.class"));
		classFiles.add( new File(pathToAllPeaClasses + "gui/kdf_setting/NumberOrEmptyFormatter.class"));
		
		classFiles.add( new File(pathToCryptoClasses + "RandomStuff.class"));
		classFiles.add( new File(pathToCryptoClasses + "AuthenticatedEncryption.class"));	
		classFiles.add( new File(pathToCryptoClasses + "HashStuff.class"));
		classFiles.add( new File(pathToCryptoClasses + "CipherStuff.class"));
		classFiles.add( new File(pathToCryptoClasses + "SessionKeyCrypt.class"));
		classFiles.add( new File(pathToCryptoClasses + "AlgoParamHandler.class"));
		classFiles.add( new File(pathToCryptoClasses + "KeyFileHandler.class"));	
		
		classFiles.add( new File(pathToCryptoClasses + "kdf" + sep + "KeyDerivation.class"));
		
		classFiles.add( new File(pathToToolsClasses + "ReadResources.class"));	
		classFiles.add( new File(pathToToolsClasses + "WriteResources.class"));	
		classFiles.add( new File(pathToToolsClasses + "FileTools.class"));
		classFiles.add( new File(pathToToolsClasses + "WebTools.class"));	
		classFiles.add( new File(pathToToolsClasses + "Converter.class"));
		classFiles.add( new File(pathToToolsClasses + "Zeroizer.class"));
		classFiles.add( new File(pathToToolsClasses + "Comparator.class"));
		classFiles.add( new File(pathToToolsClasses + "MouseRandomCollector.class"));
		classFiles.add( new File(pathToToolsClasses + "KeyRandomCollector.class"));
		classFiles.add( new File(pathToToolsClasses + "PasswordQualityCheck.class"));
		classFiles.add( new File(pathToToolsClasses + "EntropyPool.class"));
		classFiles.add( new File(pathToToolsClasses + "EntropyPool$EntropyThread.class")); // inner class
		classFiles.add( new File(pathToToolsClasses + "UnexpectedValueException.class")); // inner class
		classFiles.add( new File(pathToToolsClasses + "ZipStuff.class"));
		classFiles.add( new File(pathToToolsClasses + "ZipStuff$1.class"));	
		classFiles.add( new File(pathToToolsClasses + "ZipStuff$2.class"));	
		classFiles.add( new File(pathToToolsClasses + "ZipStuff$3.class"));	
		classFiles.add( new File(pathToToolsClasses + "PropertyHandler.class"));
		classFiles.add( new File(pathToToolsClasses + "FileDropHandler.class"));
		classFiles.add( new File(pathToToolsClasses + "ExceptionHelper.class"));
		classFiles.add( new File(pathToToolsClasses + "TestLog.class"));
		classFiles.add( new File(pathToToolsClasses + "StringNumberHelper.class")); 
		
		// CLOUD
		classFiles.add( new File(pathToAllPeaClasses + "cloud/StorageProvider.class"));	
		classFiles.add( new File(pathToAllPeaClasses + "cloud/CloudPropertyHandler.class"));
		classFiles.add( new File(pathToAllPeaClasses + "cloud/CloudFileModel.class"));
		classFiles.add( new File(pathToAllPeaClasses + "cloud/CloudControl.class"));
		classFiles.add( new File(pathToAllPeaClasses + "cloud/gui/CloudFileManager.class"));
		classFiles.add( new File(pathToAllPeaClasses + "cloud/gui/WebDavLogin.class"));
		classFiles.add( new File(pathToAllPeaClasses + "cloud/gui/WebDavLogin$WebDavLoginWorker.class"));
		classFiles.add( new File(pathToAllPeaClasses + "cloud/gui/WebDavLogin$1.class")); // WebDavLoginWorker
		classFiles.add( new File(pathToAllPeaClasses + "cloud/gui/WebDavLogin$2.class")); // SPACE action for combo box

		classFiles.add( new File(pathToAllPeaClasses + "cloud/provider/WebDavProvider.class"));
		classFiles.add( new File(pathToAllPeaClasses + "cloud/provider/WebDavProvider$1.class"));// Executer authorize
		classFiles.add( new File(pathToAllPeaClasses + "cloud/provider/WebDavProvider$2.class"));// Executer download
		classFiles.add( new File(pathToAllPeaClasses + "cloud/provider/WebDavProvider$3.class"));// Executer updateAuthorization
		classFiles.add( new File(pathToAllPeaClasses + "cloud/gui/UploadWorker.class"));
		classFiles.add( new File(pathToAllPeaClasses + "cloud/gui/UploadWorker$1.class"));
		classFiles.add( new File(pathToAllPeaClasses + "cloud/gui/DownloadWorker.class"));
		classFiles.add( new File(pathToAllPeaClasses + "cloud/gui/DownloadWorker$1.class"));
		classFiles.add( new File(pathToAllPeaClasses + "cloud/gui/CloudMenu.class"));
		
		
		//===========================
		// bouncycastle:
		classFiles.add( new File(pathToBcClasses + "util/Arrays.class"));
		classFiles.add( new File(pathToBcClasses + "util/Objects.class"));
		classFiles.add( new File(pathToBcClasses + "util/Memoable.class"));
		classFiles.add( new File(pathToBcClasses + "util/Pack.class"));
		classFiles.add( new File(pathToBcClasses + "util/Integers.class"));
		classFiles.add( new File(pathToBcClasses + "util/Longs.class"));
		classFiles.add( new File(pathToBcClasses + "util/Strings.class"));
		classFiles.add( new File(pathToBcClasses + "util/Strings$1.class"));//class StringListImpl
		
		classFiles.add( new File(pathToBcClasses + "crypto/InvalidCipherTextException.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/DataLengthException.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/OutputLengthException.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/RuntimeCryptoException.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/CryptoException.class"));
		
		classFiles.add( new File(pathToBcClasses + "crypto/BlockCipher.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/BufferedBlockCipher.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/CipherParameters.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/Digest.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/ExtendedDigest.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/Mac.class"));		
		classFiles.add( new File(pathToBcClasses + "crypto/DerivationParameters.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/DerivationFunction.class"));
		
		classFiles.add( new File(pathToBcClasses + "crypto/SkippingStreamCipher.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/StreamCipher.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/StreamBlockCipher.class"));	
		classFiles.add( new File(pathToBcClasses + "crypto/SkippingCipher.class"));
		
		classFiles.add( new File(pathToBcClasses + "crypto/macs/CMac.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/macs/HMac.class"));
		
		classFiles.add( new File(pathToBcClasses + "crypto/paddings/BlockCipherPadding.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/paddings/PaddedBufferedBlockCipher.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/paddings/PKCS7Padding.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/paddings/ISO7816d4Padding.class"));
		
		classFiles.add( new File(pathToBcClasses + "crypto/modes/CBCBlockCipher.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/modes/EAXBlockCipher.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/modes/AEADBlockCipher.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/modes/AEADCipher.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/modes/SICBlockCipher.class"));

		classFiles.add( new File(pathToBcClasses + "crypto/params/KeyParameter.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/params/ParametersWithIV.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/params/ParametersWithRandom.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/params/AEADParameters.class"));
		classFiles.add( new File(pathToBcClasses + "crypto/params/HKDFParameters.class"));		
		classFiles.add( new File(pathToBcClasses + "crypto/generators/HKDFBytesGenerator.class"));

		//=======================================
		// Variable classes: 
		//=======================================
		
		//===========================

		// Classes for all PEAs: 
			classFiles.add( new File(pathToCryptoClasses + "kdf" + sep + "CatenaKDF.class"));
			// CatenaKDF selects between all instances, so to avoid NoClassDefFoundError
			// all classes are included
			classFiles.add( new File(pathToEckCryptoClasses + "Catena.class"));
			classFiles.add( new File(pathToEckCryptoClasses + "CatenaBRG.class"));
			classFiles.add( new File(pathToEckCryptoClasses + "CatenaDBG.class"));
			classFiles.add( new File(pathToEckCryptoClasses + "CatenaHelper.class"));
			classFiles.add( new File(pathToEckCryptoClasses + "PasswordHashingScheme.class"));

			classFiles.add( new File(pathToEckCryptoClasses + "Blake2b_1.class"));				
			classFiles.add( new File(pathToEckCryptoClasses + "FastHash.class"));
			
			classFiles.add( new File(pathToCryptoClasses + "kdf" + sep + "Argon2KDF.class"));			
			classFiles.add( new File(pathToBcClasses + "crypto/generators/Argon2BytesGenerator.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/generators/Argon2BytesGenerator$Block.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/generators/Argon2BytesGenerator$FillBlock.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/generators/Argon2BytesGenerator$Position.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/params/Argon2Parameters.class"));			
			classFiles.add( new File(pathToBcClasses + "crypto/params/Argon2Parameters$Builder.class"));
			
			classFiles.add( new File(pathToBcClasses + "crypto/PasswordConverter.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/PasswordConverter$1.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/PasswordConverter$2.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/PasswordConverter$3.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/CharToByteConverter.class"));
			
			classFiles.add( new File(pathToBcClasses + "util/encoders/HexEncoder.class"));
			classFiles.add( new File(pathToBcClasses + "util/encoders/Hex.class"));
			classFiles.add( new File(pathToBcClasses + "util/encoders/EncoderException.class"));
			classFiles.add( new File(pathToBcClasses + "util/encoders/DecoderException.class"));
			classFiles.add( new File(pathToBcClasses + "util/encoders/Encoder.class"));
			classFiles.add( new File(pathToBcClasses + "util/encoders/UTF8.class"));
		
			classFiles.add( new File(pathToBcClasses + "crypto/engines/AESEngine.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/engines/SerpentEngine.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/engines/SerpentEngineBase.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/engines/ThreefishEngine.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/params/TweakableBlockCipherParameters.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/engines/ThreefishEngine$ThreefishCipher.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/engines/ThreefishEngine$Threefish1024Cipher.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/engines/ThreefishEngine$Threefish512Cipher.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/engines/ThreefishEngine$Threefish256Cipher.class"));
			
			classFiles.add( new File(pathToBcClasses + "crypto/digests/SHA3Digest.class"));	
			classFiles.add( new File(pathToBcClasses + "crypto/digests/KeccakDigest.class"));	
			classFiles.add( new File(pathToBcClasses + "crypto/digests/Blake2bDigest.class"));	
			classFiles.add( new File(pathToBcClasses + "crypto/digests/SHA512Digest.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/digests/LongDigest.class")); // for SHA-512
			classFiles.add( new File(pathToBcClasses + "crypto/digests/EncodableDigest.class")); // for SHA-512
			
			// Extra KDFs:
			if (KeyDerivation.getKdf() instanceof ScryptKDF) {				
				classFiles.add( new File(pathToCryptoClasses + "kdf" + sep + "ScryptKDF.class"));
				classFiles.add( new File(pathToBcClasses + "crypto/generators/SCrypt.class"));			
				classFiles.add( new File(pathToBcClasses + "crypto/PBEParametersGenerator.class"));
				classFiles.add( new File(pathToBcClasses + "crypto/generators/PKCS5S2ParametersGenerator.class"));		
				classFiles.add( new File(pathToBcClasses + "crypto/digests/SHA1Digest.class"));		
				classFiles.add( new File(pathToBcClasses + "crypto/digests/SHA256Digest.class"));
				classFiles.add( new File(pathToBcClasses + "crypto/digests/GeneralDigest.class"));
				classFiles.add( new File(pathToBcClasses + "crypto/engines/Salsa20Engine.class"));
				classFiles.add( new File(pathToBcClasses + "crypto/MaxBytesExceededException.class"));			
				classFiles.add( new File(pathToBcClasses + "util/StringList.class"));
				classFiles.add( new File(pathToBcClasses + "util/Iterable.class"));					
			} else if (KeyDerivation.getKdf() instanceof BcryptKDF) {
				classFiles.add( new File(pathToBcClasses + "crypto/generators/BCrypt.class"));
				classFiles.add( new File(pathToCryptoClasses + "kdf" + sep + "BcryptKDF.class"));
			} else if (KeyDerivation.getKdf() instanceof PomeloKDF) {
				classFiles.add( new File(pathToCryptoClasses + "kdf" + sep + "PomeloKDF.class"));
			} 
			
			// Extra ciphers
		if (CipherStuff.getCipherAlgo() instanceof Shacal2Engine) {
			classFiles.add( new File(pathToBcClasses + "crypto/engines/Shacal2Engine.class"));
		} else if (CipherStuff.getCipherAlgo() instanceof TwofishEngine) {
			classFiles.add( new File(pathToBcClasses + "crypto/engines/TwofishEngine.class"));			
		} else if (CipherStuff.getCipherAlgo() instanceof LEAEngine) {
			classFiles.add( new File(pathToBcClasses + "crypto/engines/LEAEngine.class"));
		} else if (CipherStuff.getCipherAlgo() instanceof RC6Engine) {
			classFiles.add( new File(pathToBcClasses + "crypto/engines/RC6Engine.class"));
		} else if (CipherStuff.getCipherAlgo() instanceof DSTU7624Engine) {
			classFiles.add( new File(pathToBcClasses + "crypto/engines/DSTU7624Engine.class"));
		}
		
		if (CipherStuff.getCipherMode() instanceof EAXMode) {
			classFiles.add( new File(pathToCryptoClasses + "EAXMode.class"));	
		}
		
		// Extra digests:
		if (HashStuff.getHashAlgo() instanceof Blake3Digest) {
			classFiles.add(  new File(pathToBcClasses + "crypto/digests/Blake3Digest.class"));
			classFiles.add(  new File(pathToBcClasses + "crypto/params/Blake3Parameters.class"));
			classFiles.add(  new File(pathToBcClasses + "crypto/Xof.class"));		
		} else if (HashStuff.getHashAlgo() instanceof WhirlpoolDigest) {
			classFiles.add(  new File(pathToBcClasses + "crypto/digests/WhirlpoolDigest.class"));
		} else if (HashStuff.getHashAlgo() instanceof SkeinDigest) {
			classFiles.add( new File(pathToBcClasses + "crypto/digests/SkeinDigest.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/digests/SkeinEngine.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/digests/SkeinEngine$UBI.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/digests/SkeinEngine$UbiTweak.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/digests/SkeinEngine$Configuration.class"));
			classFiles.add( new File(pathToBcClasses + "crypto/digests/SkeinEngine$Parameter.class"));			
		}
				
		//===========================
		// pea type specific classes: 
		
		if (dataType instanceof FileType) {
			classFiles.add( new File(pathToEckClasses + "file_pea" + sep + "FileControl.class")); 
			classFiles.add( new File(pathToEckClasses + "file_pea" + sep + "FileControl$EventDispatchThread.class")); 
			classFiles.add( new File(pathToEckClasses + "file_pea" + sep + "FileControl$ShutdownHookThread.class")); // inner class
			classFiles.add( new File(pathToEckClasses + "file_pea" + sep + "LockFrameFile.class"));
			classFiles.add( new File(pathToAllPeaClasses + "gui/TimeoutSetter.class"));

		} else if (dataType instanceof ImageType) {
			classFiles.add( new File(pathToEckClasses + "image_pea" + sep + "ImageControl.class"));
			classFiles.add( new File(pathToEckClasses + "image_pea" + sep + "ImageControl$EventDispatchThread.class")); 
			classFiles.add( new File(pathToEckClasses + "image_pea" + sep + "ImageHelper.class"));
			classFiles.add( new File(pathToEckClasses + "image_pea" + sep + "LockFrameImage.class"));		
			classFiles.add( new File(pathToEckClasses + "image_pea" + sep + "ImageSetting.class"));
			classFiles.add( new File(pathToAllPeaClasses + "gui/SingleFileSelectionDialog.class"));
			classFiles.add( new File(pathToEckClasses + "image_pea" + sep + "MetadataHandler.class"));
			classFiles.add( new File(pathToEckClasses + "image_pea" + sep + "PNGMetadataHandler.class"));
			classFiles.add( new File(pathToEckClasses + "image_pea" + sep + "JPGMetadataHandler.class"));
			classFiles.add( new File(pathToEckClasses + "image_pea" + sep + "Slideshow.class"));
			classFiles.add( new File(pathToEckClasses + "image_pea" + sep + "Slideshow$NextAction.class"));
			classFiles.add( new File(pathToEckClasses + "image_pea" + sep + "Slideshow$PreviousAction.class"));
			classFiles.add( new File(pathToEckClasses + "image_pea" + sep + "Slideshow$StopAction.class"));

		} else if (dataType instanceof CalendarType) {
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "control" + sep + "CalendarConstants.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "control" + sep + "CalendarControl.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "control" + sep + "CalendarPEAControl.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "control" + sep + "CalendarPEAControl$EventDispatchThread.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "control" + sep + "CalendarSetting.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "data" + sep + "PEvent.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "data" + sep + "PTodo.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "data" + sep + "CalendarComponent.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "data" + sep + "RecurringPEvent.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "data" + sep + "PCalendar.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "data" + sep + "PCalendar$UIDComparator.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "CalendarComboBox.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "TodoPanel.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "EventSearcher.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "CalendarMenuBar.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "CalendarMenuBar$FocusOnYearSpinnerAction.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "LockFrameCalendar.class"));
			
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "calendar_view" + sep + "CalendarLayout.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "calendar_view" + sep + "CalendarView.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "calendar_view" + sep + "DayPanel.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "calendar_view" + sep + "MonthPanel.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "calendar_view" + sep + "SingleDayPanel.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "calendar_view" + sep + "WeekPanel.class"));				
			
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "event_display" + sep + "EventInfo.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "event_display" + sep + "TodoInfo.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "event_display" + sep + "ComponentInfo.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "event_display" + sep + "TentativeEventPanel.class"));	
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "event_display" + sep + "CancelledEventPanel.class"));	
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "event_display" + sep + "EventButton.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "event_display" + sep + "TodoButton.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "event_display" + sep + "ComponentButton.class"));
			
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "event_setting" + sep + "DatePicker.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "event_setting" + sep + "EventSetter.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "event_setting" + sep + "TodoSetter.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "event_setting" + sep + "ComponentSetter.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "event_setting" + sep + "RecurrencePanel.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "event_setting" + sep + "TimePicker.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "event_setting" + sep + "ExceptionPanel.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "event_setting" + sep + "ByDayRecurrencePanel.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "gui" + sep + "event_setting" + sep + "AtRestDialog.class"));
									
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "util" + sep + "ReadICal.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "util" + sep + "WriteICal.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "util" + sep + "ICalDateTimeHelper.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "util" + sep + "ICalRecurrenceHelper.class"));			
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "util" + sep + "CalendarComposer.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "util" + sep + "ICalStringHelper.class"));
			classFiles.add( new File(pathToEckClasses + "calendar_pea" + sep + "util" + sep + "RecurringEventComposer.class"));

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! TODO TODO TODO 
//			classFiles.add( new File(pathToEckClasses + "js_cal" + sep + "DatenSetzen.class"));
//			classFiles.add( new File(pathToEckClasses + "js_cal" + sep + "JS_Setter.class"));
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! TODO TODO TODO 
		} else if (dataType instanceof EditorType) {
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "EditorControl.class"));
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "EditorControl$EventDispatchThread.class")); 
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "LockFrameEditorMenu.class"));
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "TextHelper$TxtFilter.class")); // inner class
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "TextHelper$RtfFilter.class")); // inner class
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "TextHelper$HTMLFilter.class")); // inner class
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "TextHelper$RtfHtmlFilter.class")); // inner class
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "LockFrameEditor.class"));
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "CopyCutPastePopup.class"));
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "LockFrameEditor$UndoAction.class")); // inner class
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "LockFrameEditor$RedoAction.class")); // inner class
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "LockFrameEditor$ReturnToTextAction.class"));			
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "TabComponent.class"));
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "TabComponent$BackgroundPainter.class")); // Painter for JTextPane background
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "TabComponent$GoToTabAction.class"));
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "NotebookSetting.class"));
			//classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "NotebookSetting$1.class"));// flavor listener
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "PeaTextPane.class"));
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "NotebookModel.class"));
			classFiles.add( new File(pathToEckClasses + "note_pea" + sep + "TextHelper.class"));			
			classFiles.add( new File(pathToAllPeaClasses + "gui/PasswordGeneratorDialog.class"));
		
			
			//
			// Add library files to jar file
			// 
/*			String pathLibraryClasses = "bin" + sep + "com" + sep;
			File libs = new File(pathLibraryClasses);
			if (! libs.exists()) {
				TestLog.e(XXXX.class, "Can't find libraries...");
				System.exit(1);
			}
			ArrayList<File> libClassNamelist = FileTools.listAllFilesOfFolder(libs, true);
			//String[] libClassNames = new String[libClassNamelist.size()];
			//libClassNames = libClassNamelist.toArray(libClassNames);
			classFiles.addAll(libClassNamelist);
*/			
			
		} else {
			TestLog.e(CreateJarFile.class, "CreateJarFile: Invalid file type.");
			System.exit(3);
		}

		//
		// stream to write jar file:
		//
		JarOutputStream jos = null;
		try {
			// create jar file with manifest			
			jos = new JarOutputStream(new FileOutputStream(jarFile), manifest);
		} catch (FileNotFoundException e) {
			TestLog.e(CreateJarFile.class, e.getLocalizedMessage() + " - " + e.toString(), 5);
		} catch (IOException e) {
			TestLog.e(CreateJarFile.class, e.getLocalizedMessage() + " - " + e.toString(), 5);
		}

		//
		// Add class files to jar file
		//
		for ( File source : classFiles ) {		
			BufferedInputStream in = null;
			try  {
				// cut "bin/" and replace File.sep with "/"
				JarEntry entry = null;
				if(source.getPath().startsWith("settings")
						|| source.getPath().startsWith("start")){
					entry = new JarEntry( (source.getPath().replace(File.separator, "/") ));	
				} else {
					entry = new JarEntry( (source.getPath().substring(4, source.getPath().length())).replace(File.separator, "/") );	
				}
				entry.setTime(source.lastModified());
				jos.putNextEntry(entry);
				in = new BufferedInputStream(new FileInputStream(source));
			    byte[] buffer = new byte[1024];
			    while (true)  {
			    	int count = in.read(buffer);
			    	if (count == -1)  break;
			    	jos.write(buffer, 0, count);
			    }
			    jos.closeEntry();
			} catch (IOException e) {
				TestLog.e(CreateJarFile.class, e.getLocalizedMessage() + " - " + e.toString(), 5);
			}  finally {
				if (in != null) {
					try {
						in.close();
					} catch (IOException e) {
						TestLog.e(CreateJarFile.class, e.getLocalizedMessage() + " - " + e.toString(), 5);
					}
				}
			}
		} // End for classFiles
				

		//
		// Add resource files to jar file
		//
		// from directory "resources"
		//
		File resourcesDir = new File("resources");

		if (! resourcesDir.exists() ) TestLog.e(CreateJarFile.class, "Error: resources does not exist: " + resourcesDir.getAbsolutePath() );
		if (! resourcesDir.isDirectory() ) TestLog.e(CreateJarFile.class, "Error: resources is not directory");
		if ( resourcesDir.listFiles().length == 0 ) TestLog.e(CreateJarFile.class, "Error: resources is empty");

		for ( File source : resourcesDir.listFiles() ) {	
			String fileName = source.getName();

			if (source.getName().equals("pea-lock.png") // image icon
					|| fileName.equals("wait_cursor.png") // image icon
					|| fileName.equals("back24.png") // image icon
					|| fileName.equals("groesser.png") // image icon
					|| fileName.equals("kleiner.png") // image icon
					|| fileName.equals("weiter.png") // image icon
					|| fileName.equals("zurueck.png") // image icon
					|| fileName.equals("update.png") // image icon
					|| fileName.equals("eye.png") // image icon
					|| fileName.equals("info.png") // image icon
					|| fileName.equals("locked.png") // image icon
					|| fileName.equals("unlocked.png") // image icon
					|| fileName.equals("PeaLanguagesBundle.properties") // base bundle: include always
					|| fileName.equals("pea.properties") // properties for all PEAs
					|| fileName.equals("first_time_guide.html") // guide en
					|| fileName.equals("cloud_guide.html") // guide en
					|| ((dataType instanceof EditorType)
							&& fileName.equals("center-align.png"))
					|| ((dataType instanceof EditorType)
							&& fileName.equals("center-align-rollover.png"))
					|| ((dataType instanceof EditorType)
							&& fileName.equals("left-align.png"))
					|| ((dataType instanceof EditorType)
							&& fileName.equals("left-align-rollover.png"))
					|| ((dataType instanceof EditorType)
							&& fileName.equals("right-align.png"))
					|| ((dataType instanceof EditorType)
							&& fileName.equals("right-align-rollover.png"))
					|| ((dataType instanceof EditorType)
							&& fileName.equals("colors.png"))
					
					|| (fileName.equals("date-picker.png") && dataType instanceof CalendarType)
					|| ((dataType instanceof CalendarType)
							&& fileName.equals("CalendarLanguageBundle.properties")) // base bundle: include always
					// PEA specific properties:
					|| ((dataType instanceof CalendarType)
							&& fileName.equals("calendar.properties"))
					|| ((dataType instanceof EditorType)
							&& fileName.equals("editor.properties"))
					|| ((dataType instanceof FileType)
							&& fileName.equals("file.properties"))
					|| ((dataType instanceof ImageType)
							&& fileName.equals("image.properties"))
					) {				
				if (fileName.equals("pea.properties")){
					// delete lastSalt
					String error = new PropertyHandler().deleteLastSaltProperty();
					if (error != null) {
						TestLog.e(CreateJarFile.class, "Could not remove lastSalt property: " + error);
					}
				}

				BufferedInputStream in = null;
				try  {
					JarEntry entry = new JarEntry(source.getPath().replace(File.separator, "/"));

					entry.setTime(source.lastModified());
					jos.putNextEntry(entry);
					in = new BufferedInputStream(new FileInputStream(source));
					byte[] buffer = new byte[1024];
					while (true)  {
						int count = in.read(buffer);
						if (count == -1)  break;
						jos.write(buffer, 0, count);
					}
					jos.closeEntry();
				} catch (IOException e) {
					TestLog.e(CreateJarFile.class, e.getMessage() + " - " + e.toString(), 5);
				}  finally {
					if (in != null) {
						try {
							in.close();
						} catch (IOException e) {
							TestLog.e(CreateJarFile.class, e.getMessage() + " - " + e.toString(), 5);
						}
					}
				}
			} else {
				
			}
		} // End for resources
		
		// 
		// Add languages: sub-directory of resources (country code)
		// 
		if (langsToAdd != null && langsToAdd.isEmpty() == false) {
			for (String lang : langsToAdd) {
				File langDir = new File(resourcesDir + File.separator + lang);
				if ( ! langDir.exists() || langDir.isDirectory() == false) {
					TestLog.ve(CreateJarFile.class, "Folder for language in resources not found: " + lang, 5);
					continue;
				}
				for ( File source : langDir.listFiles() ) {	
					String fileName = source.getName();
					if (source.getName().startsWith("first_time_guide") 
							|| fileName.startsWith("cloud_guide") 
							|| fileName.startsWith("PeaLanguagesBundle") 
							|| ( dataType instanceof CalendarType
									&& fileName.startsWith("CalendarLanguageBundle") ) ) {
						BufferedInputStream in = null;
						try  {
							JarEntry entry = new JarEntry(source.getPath().replace(File.separator, "/"));

							entry.setTime(source.lastModified());
							jos.putNextEntry(entry);
							in = new BufferedInputStream(new FileInputStream(source));
							byte[] buffer = new byte[1024];
							while (true)  {
								int count = in.read(buffer);
								if (count == -1)  break;
								jos.write(buffer, 0, count);
							}
							jos.closeEntry();
							TestLog.v(CreateJarFile.class, "Added language: " + lang);
						} catch (IOException e) {
							TestLog.e(CreateJarFile.class, e.getMessage() + " - " + e.toString(), 5);
						}  finally {
							if (in != null) {
								try {
									in.close();
								} catch (IOException e) {
									TestLog.e(CreateJarFile.class, e.getMessage() + " - " + e.toString(), 5);
								}
							}
						}						
					} 
				}
			}
		} else {
			TestLog.v(CreateJarFile.class, "No language to add");
		}
		
		
		//
		// Add README.txt files to jar file
		//
		File readme = null;
		if (dataType instanceof EditorType) {
			
			try {
				Files.copy(Paths.get("resources" + File.separator + "notes" + File.separator + "README.txt"), 
						Paths.get("README.txt"), StandardCopyOption.REPLACE_EXISTING);
			} catch (IOException e) {
				TestLog.e(CreateJarFile.class, e.getLocalizedMessage() + " - " + e.toString(), 5);
			}
			readme = new File("README.txt");
		} else if (dataType instanceof ImageType) {
			
			try {
				Files.copy(Paths.get("resources" + File.separator + "image" + File.separator + "README.txt"), 
						Paths.get("README.txt"), StandardCopyOption.REPLACE_EXISTING);
			} catch (IOException e) {
				TestLog.e(CreateJarFile.class, e.getLocalizedMessage() + " - " + e.toString(), 5);
			}
			readme = new File("README.txt");
		} else if (dataType instanceof CalendarType) {
			
			try {
				Files.copy(Paths.get("resources" + File.separator + "Calendar" + File.separator + "README.txt"), 
						Paths.get("README.txt"), StandardCopyOption.REPLACE_EXISTING);
			} catch (IOException e) {
				TestLog.e(CreateJarFile.class, e.getLocalizedMessage() + " - " + e.toString(), 5);
			}
			readme = new File("README.txt");
		} else if (dataType instanceof FileType) {
			
			try {
				Files.copy(Paths.get("resources" + File.separator + "File" + File.separator + "README.txt"), 
						Paths.get("README.txt"), StandardCopyOption.REPLACE_EXISTING);
			} catch (IOException e) {
				TestLog.e(CreateJarFile.class, e.getLocalizedMessage() + " - " + e.toString(), 5);
			}
			readme = new File("README.txt");
		}
		if (readme == null) {
			new UnexpectedValueException("File", "readme", "is null").printDescription();
		}
		if (! readme.exists() ) TestLog.e(CreateJarFile.class, "Error: README does not exist: " + resourcesDir.getAbsolutePath() );
		if (! readme.isFile() ) TestLog.e(CreateJarFile.class, "Error: README is not a file");


		BufferedInputStream in = null;
		try  {
			JarEntry entry = new JarEntry(readme.getPath().replace(File.separator, "/"));

			entry.setTime(readme.lastModified());
			jos.putNextEntry(entry);
			in = new BufferedInputStream(new FileInputStream(readme));
			byte[] buffer = new byte[1024];
			while (true)  {
				int count = in.read(buffer);
				if (count == -1)  break;
				jos.write(buffer, 0, count);
			}
			jos.closeEntry();
		} catch (IOException e) {
			TestLog.e(CreateJarFile.class, e.getLocalizedMessage() + " - " + e.toString(), 5);
		}  finally {
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
					TestLog.e(CreateJarFile.class, e.getLocalizedMessage() + " - " + e.toString(), 5);
				}
			}
		}
		
		
		//Add class files from otherlibraries: sardine, apache.org
		//String pathToSardine = "bin" + sep + "com" + sep + "github" + sep 
		//		+ "sardine";
		String pathToSardine = "bin" + sep + "com" + sep + "thegrizzlylabs" + sep 
				+ "sardineandroid";
		File sardineDir = new File(pathToSardine);

		if (! sardineDir.exists() ) TestLog.e(CreateJarFile.class, "Error: sardineDir does not exist: " + sardineDir.getAbsolutePath() );
		if (! sardineDir.isDirectory() ) TestLog.e(CreateJarFile.class, "Error: sardineDir is not directory");
		if ( sardineDir.listFiles().length == 0 ) TestLog.e(CreateJarFile.class, "Error: sardineDir is empty");
		
		addAllEntriesOfDir(in, sardineDir, jos);
		
		String pathToOkHttp = "bin" + sep + "okhttp3";
		File okHttpDir = new File(pathToOkHttp);

		if (! okHttpDir.exists() ) TestLog.e(CreateJarFile.class, "Error: okhttp3 does not exist: " + okHttpDir.getAbsolutePath() );
		if (! okHttpDir.isDirectory() ) TestLog.e(CreateJarFile.class, "Error: okhttp3 is not directory");
		if ( okHttpDir.listFiles().length == 0 ) TestLog.e(CreateJarFile.class, "Error: okhttp3 is empty");
		
		addAllEntriesOfDir(in, okHttpDir, jos); 
		
		String pathToOkio = "bin" + sep + "okio";
		File okioDir = new File(pathToOkio);

		if (! okioDir.exists() ) TestLog.e(CreateJarFile.class, "Error: okio does not exist: " + okioDir.getAbsolutePath() );
		if (! okioDir.isDirectory() ) TestLog.e(CreateJarFile.class, "Error: okio is not directory");
		if ( okioDir.listFiles().length == 0 ) TestLog.e(CreateJarFile.class, "Error: okio is empty");
		
		addAllEntriesOfDir(in, okioDir, jos); 		
		
		String pathToSimpleXML = "bin" + sep + "org" + sep + "simpleframework";
		File simpleXmlDir = new File(pathToSimpleXML);

		if (! simpleXmlDir.exists() ) TestLog.e(CreateJarFile.class, "Error: simpleframework does not exist: " + simpleXmlDir.getAbsolutePath() );
		if (! simpleXmlDir.isDirectory() ) TestLog.e(CreateJarFile.class, "Error: simpleframework is not directory");
		if ( simpleXmlDir.listFiles().length == 0 ) TestLog.e(CreateJarFile.class, "Error: simpleframework is empty");
		
		addAllEntriesOfDir(in, simpleXmlDir, jos); 
		
		//
		// Add library files to jar file
		//
		// from directory "peafactory_lib"
		//
/*		File lib = new File("peafactory_lib");
 * needed: e.g. http://www.jdotsoft.com/JarClassLoader.php

		if (! lib.exists() ) TestLog.e(XXXX.class, "Error: peafactory_lib does not exist: " + resourcesDir.getAbsolutePath() );
		if (! lib.isDirectory() ) TestLog.e(XXXX.class, "Error: lib is not directory");
		if ( lib.listFiles().length == 0 ) TestLog.e(XXXX.class, "Error: lib is empty");
		
		for ( File source : lib.listFiles() ) {		
			// + File.separator + "dropbox-core-sdk-3.1.1" + File.separator + "dropbox-core-sdk-3.1.1.jar");
			if (source.getName().endsWith("okhttp-3.14.0.jar") // okhtto
					|| source.getName().endsWith("okio-1.17.5.jar") // okio used in okhttp
					|| source.getName().endsWith("simple-xml-2.7.1.jar") // simple xml used in okhttp and sardineandroid


					) {				
				BufferedInputStream libIn = null;
				String sourcePath = source.getPath();
	//			sourcePath = sourcePath.substring(sourcePath.indexOf(File.separator), sourcePath.length());
				sourcePath.replace(File.separator, "/");
				try  {
					JarEntry entry = new JarEntry(sourcePath);//source.getPath().replace(File.separator, "/"));

					entry.setTime(source.lastModified());
					jos.putNextEntry(entry);
					libIn = new BufferedInputStream(new FileInputStream(source));
				    byte[] buffer = new byte[1024];
				    while (true)  {
				    	int count = libIn.read(buffer);
				    	if (count == -1)  break;
				    	jos.write(buffer, 0, count);
				    }
				    jos.closeEntry();
				} catch (IOException e) {
					TestLog.e(CreateJarFile.class, e.getLocalizedMessage() + " - " + e.toString(), 5);
				}  finally {
					if (libIn != null) {
						try {
							libIn.close();
						} catch (IOException e) {
							TestLog.e(CreateJarFile.class, e.getLocalizedMessage() + " - " + e.toString(), 5);
						}
					}
				}
			} else {
				
			}
		} // End for libraries
*/		
		
		if (jos != null) {
			try {
				jos.flush();
				jos.close();
			} catch (IOException e) {
				TestLog.e(CreateJarFile.class, e.getLocalizedMessage() + " - " + e.toString(), 5);
			}
		} 
		// set jar file executable if possible: 
		// (does not overwrite settings from partition)
		if ( ! jarFile.setExecutable(true,true) ) { // only executable for owner: second true
			try{
				Runtime.getRuntime().exec("chmod u+x " + JarStuff.getJarFileName());
			} catch (Exception e) {			
				TestLog.e(CreateJarFile.class, "Can not set jar-archive executable.");
			}
		}
			
		if (PeaFactory.isUnixScript() == true) {
			
			// jar file without ".jar"
			String jarName = JarStuff.getJarFileName().substring(0, JarStuff.getJarFileName().length() - 4);

			try {
				// create script content:
				String scriptContent = "#!/bin/sh \n"
						+ "java -jar " + JarStuff.getJarFileName();
				
				// get file name of script
				String fileName  = "peas" + File.separator 
						+ "start_" + jarName + ".sh";				
				
				// Write script file:
				WriteResources.writeText(scriptContent, fileName);
				
				// make script file executable:
				Runtime.getRuntime().exec("chmod u+x " + fileName);
				
			} catch (Exception e) {
				TestLog.e(CreateJarFile.class, "Error: can not create unix script");
			}			
			
		} else {
			//TestLog.o(XXXX.class, "jar-archive is executable for owner.");
		}		
	}
	private static void addAllEntriesOfDir(BufferedInputStream in, File dir, JarOutputStream jos) {
		if (dir == null) {
			TestLog.e(CreateJarFile.class, "CreateJarFile.addAllEntriesOfDir: Directory is null");
			return;
		}
		if (! dir.isDirectory()) {
			new UnexpectedValueException("File", "dir", "is not directory").printDescription();
			return;
		}
		File[] fileList = dir.listFiles();
		if (fileList == null) {
			new UnexpectedValueException("File[]", "fileList", "is null for: "+ dir.getAbsolutePath()).printDescription();
			return;
		}
		for ( File source : fileList ) {
			if (source.isDirectory()) {
				if (source.getName().equals("class-use")) {
					// do nothing
				} else {
					addAllEntriesOfDir(in, source, jos);
				}
			} else {
				if (source.getName().endsWith("html")) {
					// do nothing
				} else {
					JarEntry entry = new JarEntry( (source.getPath().substring(4, source.getPath().length())).replace(File.separator, "/") );	
					entry.setTime(source.lastModified());
					try {
						jos.putNextEntry(entry);
						in = new BufferedInputStream(new FileInputStream(source));
						byte[] buffer = new byte[1024];
						while (true)  {
							int count = in.read(buffer);
							if (count == -1)  break;
							jos.write(buffer, 0, count);
						}
						jos.closeEntry();
					} catch (IOException e) {
						TestLog.e(CreateJarFile.class, new ExceptionHelper(e).getInfos());
					}
				}
			}
		}
	}
	/**
	 * Add a language bundle to the PEA to create
	 * 
	 * @param newLang	language short name e.g. "en", "tr"
	 */
	public static void addLang(String newLang) {
		if ( ! langsToAdd.contains(newLang)) {
			langsToAdd.add(newLang);
		}
	}
	/**
	 * Remove a language bundle from the PEA to create
	 * 
	 * @param langToRemove	language short name e.g. "en", "tr"
	 */
	public static void removeLang(String langToRemove) {
		langsToAdd.remove(langToRemove);
	}
}
