package cologne.eck.peafactory.control;

/*
 * Peafactory - Production of Password Encryption Archives
 * Copyright (C) 2015  Axel von dem Bruch
 * 
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, 
 * or (at your option) any later version.
 * This library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * See:  http://www.gnu.org/licenses/gpl-2.0.html
 * You should have received a copy of the GNU General Public License 
 * along with this library.
 */

/**
 * Modifies the java files with new settings and random values. 
 * These files will be compiled later by FileCompiler. 
 * Creates two files:
 *
 * - start/Start.java: to start the program with enough memory and in specified mode,
 * 		to check the required java version and to check if an instance of this
 *      program is already running (OS dependent). 
 *      Redirects out and err. 
 *      
 * - settings/PeaSettings.java: sets cryptographic primitives, type of pea, parameters
 *      of the key derivation function, random values, optional features like virtual 
 *      keyboard, password generator
 */


import java.io.*;

import javax.swing.JOptionPane;

import org.bouncycastle.crypto.BlockCipher;
import org.bouncycastle.crypto.Digest;
import org.bouncycastle.crypto.digests.Blake2bDigest;
import org.bouncycastle.crypto.digests.Blake3Digest;
import org.bouncycastle.crypto.digests.SHA3Digest;
import org.bouncycastle.crypto.digests.SHA512Digest;
import org.bouncycastle.crypto.digests.SkeinDigest;
import org.bouncycastle.crypto.digests.WhirlpoolDigest;
import org.bouncycastle.crypto.engines.AESEngine;
import org.bouncycastle.crypto.engines.DSTU7624Engine;
import org.bouncycastle.crypto.engines.LEAEngine;
import org.bouncycastle.crypto.engines.RC6Engine;
import org.bouncycastle.crypto.engines.SerpentEngine;
import org.bouncycastle.crypto.engines.Shacal2Engine;
import org.bouncycastle.crypto.engines.ThreefishEngine;
import org.bouncycastle.crypto.engines.TwofishEngine;

import cologne.eck.all_peas.data.Attachments;
import cologne.eck.all_peas.data.PeaProperties;
import cologne.eck.calendar_pea.control.CalendarPEAControl;
import cologne.eck.file_pea.FileControl;
import cologne.eck.image_pea.ImageControl;
import cologne.eck.note_pea.EditorControl;
import cologne.eck.peafactory.control.data_type.CalendarType;
import cologne.eck.peafactory.control.data_type.DataType;
import cologne.eck.peafactory.control.data_type.EditorType;
import cologne.eck.peafactory.control.data_type.FileType;
import cologne.eck.peafactory.control.data_type.ImageType;
import cologne.eck.peafactory.crypto.CipherStuff;
import cologne.eck.peafactory.crypto.HashStuff;
import cologne.eck.peafactory.crypto.kdf.*;
import cologne.eck.tools.TestLog;


class FileModifier {
		

	/**
	 * Generate the file "start/Start.java". 
	 * This file will be compiled and stored in the new PEA
	 */
	protected final static void generateStart(){
		
		StringBuilder builder = new StringBuilder();
		
		// Fixed content: 
		builder.append(
				
				"package start;\n"
				+"\n/* Automatically produced file */\n"
				+ "/* This class is newly created for each PEA */\n\n"
				+"import java.io.*;\n"
				+"import javax.swing.JOptionPane;\n"
				+ "import cologne.eck.all_peas.data.*;\n"
				+ "import cologne.eck.all_peas.vm_specific.JREBridge;\n"
				+ "import cologne.eck.tools.Zeroizer;\n\n"

				+"public class Start {\n\n"
				);

		
		// variable content: 
		String base = "private static final String CLASS_NAME = ";
		if (DataType.getCurrentType() instanceof FileType ) {
			builder.append(base + "\"file_pea/FileControl\";\n");	
		} else if (DataType.getCurrentType() instanceof ImageType ) {
			builder.append(base + "\"image_pea/ImageControl\";\n");						
		} else if (DataType.getCurrentType() instanceof EditorType ) {
			builder.append(base + "\"note_pea/EditorControl\";\n");								
		} else if (DataType.getCurrentType() instanceof CalendarType ) {
			builder.append(base + "\"calendar_pea/control/CalendarPEAControl\";\n");								
		} 
		if (JarStuff.getJarFileName() != null) {
			if (JarStuff.getJarFileName().contains("\"")){ // avoid executable code...
				TestLog.e(FileModifier.class, "Invalid jar file name: contains quotes");
				PeaProperties.getVmBridge().exitWithError(null, "FileModifier: Invalid jar file name: contains quotes", false);
			}			
			builder.append("private static final String JAR_FILE_NAME = "
				+ "\"" + JarStuff.getJarFileName().substring(0, JarStuff.getJarFileName().length() -4) + "\";\n");
		} else {
			builder.append("private static final String JAR_FILE_NAME = \"default\";\n");			
		}
		
		// Fixed content:
		builder.append(
				"private static int requiredJavaVersion = 8;\n"
			+	"private static Runtime run = Runtime.getRuntime();\n\n"

+				"public static void main(String[] args) {\n\n"


+ 				"   System.out.println(\"Starting PEA \" + JAR_FILE_NAME +\" by new process...\");\n"

// Check java version:
+				"   // check java version:\n"
+				"   // Show error dialog only, when version is definitely to low \n"
+               "   PeaProperties.setVmBridge(new JREBridge());\n"
+				"   String jvString = null;\n"
+				"   try {\n"
+				"      jvString = System.getProperty(\"java.version\");\n"
+				"   } catch (Exception e) {\n"
+				"      System.err.println( \"Could not check java version... \" );\n"
+				"   }\n\n"
+				"  try {\n"
+				"     int javaVersion = 0;\n"
+				"     if (jvString.contains(\".\")) { //old format 1.X or Windows e.g. 16.0.1 \n"
+				"       String[] splitted = jvString.split(\"\\\\.\");\n"
+				"       if (splitted.length > 1) {\n"
+				"         if (Integer.parseInt(splitted[0]) == 1) { // old version 1.X.X \n"
+				"           javaVersion = Integer.parseInt(splitted[1]);// get second \n"
+				"         } else {  \n"
+				"         javaVersion = Integer.parseInt(splitted[0]); \n"
+				"       } \n"
+				"    } else { \n"
+				"       javaVersion = Integer.parseInt(jvString); \n"
+				"    } \n"
+				"  } else { // new format from from java 9 \n"
+				"    javaVersion = Integer.parseInt(jvString); \n"
+				"  }\n"
+				"  if (javaVersion < requiredJavaVersion) { \n"
+				"    JOptionPane.showMessageDialog(null,\n"
+				"    \"Wrong java version. Required java version: \" + requiredJavaVersion + \", available: \" + javaVersion  + \"   ( \" + System.getProperty(\"java.version\") +  \" ) \",\n"
+				"    null, JOptionPane.ERROR_MESSAGE);\n"
+				"    System.exit(1); }\n"
+				"} catch (Exception e) {\n"
+				"   System.err.println( \"Could not check java version: \" + System.getProperty(\"java.version\"));\n"
+				"   System.err.println( \"Warning: The used version may be inappropriate...\" + jvString);\n"
+				"}\n\n"

// check running program
+				"// check if there is an instance of this program already running:\n"
+				"boolean warn = false;\n"
+				"// check for windows and unix:\n"
+				"String OS = System.getProperty(\"os.name\").toLowerCase();\n\n"
+				"String procs = \"\";\n"
+				"if (OS.contains(\"windows\")){\n"
+				"  String line;\n"
+				"  try { // read processes, look for this PEA \n"
+			    "    Process procCheckWin = Runtime.getRuntime().exec(\"wmic PROCESS where \\\"name like \'%java%\'\\\" get Processid,Caption,Commandline\");\n"
+			    "    BufferedReader inputCheckWin = new BufferedReader(new InputStreamReader(procCheckWin.getInputStream()));\n"
+			    "    while ((line = inputCheckWin.readLine()) != null) {\n"	
+		        "      if(line.contains(JAR_FILE_NAME + \".jar\")){\n"	
+				"        System.out.println( line);\n"
+				"        procs = procs + line + \"\\n\";\n"
+				"        if (line.contains(\"Files\\\\Oracle\\\\Java\\\\javapath\\\\\")){\n"
+				"           continue;\n"
+				"        } else {\n"
+				"           if(line.contains(\"-Xms\")){\n"//firstOccurence == true){// process is running already\n"
+        		"             warn = true;\n"
+				"             System.out.println( \"instance already running\");\n"
+	        	"             break;\n"
+				"           } else {\n"
+		        "           }\n"
+				"        }\n"
+ 				"      }\n"
+			    "    }\n"
+			    "    inputCheckWin.close();\n"
+				"  } catch (IOException ioe) {\n"
+					 "JOptionPane.showMessageDialog(null, ioe.toString() + \" - check process failed\");\n"
+			    "    ioe.printStackTrace();\n"
+				"}\n"
+				"} else { // should work for most Unix: Linux, BSD, Solaris, Mac OS except OS X \n"
+				"  try {//  \n"
+				"    Process checkPro = run.exec(\"pgrep -f \"+ JAR_FILE_NAME + \".jar\");\n"
// does not work... +				"Process checkPro = run.exec(\"ps -axo pid,command,args | grep \" + JAR_FILE_NAME + \".jar | awk \'{ print $1 }\'\");\n"
+				"    checkPro.waitFor();\n"
+				"    BufferedReader buffCheck = new BufferedReader(new InputStreamReader(checkPro.getInputStream()));\n"
+				"    buffCheck.readLine(); // this is always existent: the ID of this process\n"
+				"    int checkProID = Integer.parseInt(buffCheck.readLine()); // next line: there should be no ID\n"
//+				"    //TestLog.o(Start.class, checkProID);\n"
+ 				"    if( checkProID > 100){\n"
+				"      System.out.println(\"\" + checkProID);\n"
+				"      procs = procs + \"Process: \" + checkProID + \"\\n\";\n"
+				"      warn = true;\n"
+				"      System.out.println( \"instance already running\");\n"
+				"       buffCheck.close();\n"
+				"    }\n"
+				"  } catch (NumberFormatException e) { \n"
+				"    // thats ok, there is no such process\n"
+				"  } catch (Exception e) { \n"
+					 "JOptionPane.showMessageDialog(null, e.getMessage() + \" - check process failed\");\n"
+				"    //... this will be thrown for rarely used systems with no pgrep\n"
+				"  }\n"
+				"}\n\n"

// warn if the program is already running:
+				"if(warn == true){ \n"
+				"  Object[] options = {PeaProperties.getVmBridge().translate(\"start_nevertheless\"), PeaProperties.getVmBridge().translate(\"cancel\")};\n"
+				"  int n = JOptionPane.showOptionDialog(null,\n"
+				"    PeaProperties.getVmBridge().translate(\"running_instance\") + \"\\n\" + procs,\n"
+				"    PeaProperties.getVmBridge().translate(\"warning\"), JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE, null, \n"
+			    "    options, options[1]);\n"		
+				"  if(n != 0) { System.exit(0);}\n"
+				"}\n\n"

// check if program runs in command line mode and password is not yet set:
+ "//check command line mode:\n"
+ "boolean containsFiles = false;\n"
+ "boolean containsPassword = false;\n"
+ "boolean addPassword = false;\n"
+ "char[] psw = null;\n"
+ "if (args.length > 0) {\n"
+ "	\n"
+ "	for(int i = 0; i < args.length; i++) {\n"
+ "		String s = args[i];\n"
+ "		// check if passphrase is present:\n"
+ "		if (s.equals(\"--passphrase\") || s.equals(\"--password\")) {\n"
+ "			containsPassword = true;\n"
+ "		}\n"
+ "		if (s.equals(\"--files\") ) {\n"
+ "			containsFiles = true;\n"
+ "		}\n"
+ "	}\n"
+ "	if (containsFiles == true && containsPassword == false) {		\n"
+ "		Console console = System.console();\n"
+ "		if (console != null) {\n"
+ "			try {\n"
+ "				//TestLog.o(CommandLineHandler.class, \"Enter the password:\");\n"
+ "				psw = console.readPassword(PeaProperties.getVmBridge().translate(\"enter_password\"));\n"
+ "				addPassword = true;\n"
+ "			} catch (Exception e) {\n"
+ "				System.err.println( e.getLocalizedMessage() + \" - \" + e.toString());\n"
+ "			}\n"
+ "		}\n"
+ "	}\n"
+ "}\n"
+ "\n"

// create the commands for ProcessBuilder:
+ "String[] commands = new String[args.length + 4 + ((addPassword == true)? 2 : 0)];\n"
+ "commands[0] = \"java\";\n"
+ "commands[1] = \"-cp\";\n"
+ "commands[2] = JAR_FILE_NAME + \".jar\";\n"
+ "commands[3] = \"cologne.eck.\" + CLASS_NAME;"
+ "System.arraycopy(args, 0, commands, 4, args.length);\n"
+ "if (addPassword == true) {\n"
+ "	commands[commands.length - 2] = \"--passphrase\";\n"
+ "	if (psw == null) { // null password\n"
+ "		commands[commands.length - 1] =  \"\\\"\\\"\";\n"
+ "	} else {\n"
+ "		commands[commands.length - 1] =  \"\\\"\" + new String(psw) + \"\\\"\";\n"
+ "		Zeroizer.zero(psw);\n"
+ "	}\n"
+ "}\n"
+ "\n"
//+ "String x = \"\"; for (String s : commands) { System.out.println(s);x+=s+\" \";}System.out.println(x);\n"
+ "\n"
+ ""

// start the process:
+				"try {\n"
+				"  // run the program\n"
+ "   ProcessBuilder pb = new ProcessBuilder();\n"
+ 				"   @SuppressWarnings(\"unused\")\n"
+ "Process p = pb.inheritIO().command(commands).start();\n"
+ 				"   System.exit(0);\n"
+				"} catch (IOException ie) {\n"
+				"  JOptionPane.showMessageDialog(null,\n"
+				"  \"Execution failed.\"\n"
+				"  + ie.toString(),\n"
+				"  null,JOptionPane.ERROR_MESSAGE);\n"
+				"  ie.printStackTrace();\n"
+				"  System.exit(0);\n"
+				"} catch (Exception ire) { // InterruptedException  //for terminal only\n"
+				"  JOptionPane.showMessageDialog(null,\n"
+				"  \"Execution failed.\"\n"
+				"  + ire.toString(),\n"
+				"  null,JOptionPane.ERROR_MESSAGE);\n"
+				"  ire.printStackTrace();\n"
+				"  System.exit(0);\n"
+				"} finally{"
+ 				"    System.exit(0);\n"
+ 				"}}}\n"
				);
		
		// Write file: 
		BufferedWriter writer = null;
		try {
			writer = new BufferedWriter( new FileWriter("start" + File.separator + "Start.java"));
			writer.write( new String(builder) );
			writer.close();
		} catch (IOException e) {
			TestLog.e(FileModifier.class, e.getLocalizedMessage() + " - " + e.toString(), 5);
		}
	}
	
	/**
	 * Generate the file "settings/PeaSettings.java". 
	 * This file will be compiled and stored in the new PEA
	 */
	
	protected final static void generatePeaSettings() {
		
		StringBuilder builder = new StringBuilder();
		
		// Fixed content:
		builder.append("package settings;\n"
+"\n/* Automatically produced file */\n"
+ "/* This class is newly created for each PEA */\n\n"
+ "import javax.swing.JDialog;\n"
+ "import cologne.eck.all_peas.bridge.*;\n"
+ "import cologne.eck.all_peas.vm_specific.*;\n"
+ "import org.bouncycastle.crypto.digests.*;\n"
+ "import org.bouncycastle.crypto.engines.*;\n"
+ "import org.bouncycastle.crypto.*;\n"
+ "import cologne.eck.peafactory.crypto.kdf.*;\n");
		
		// Variable content:
		builder.append("\npublic class PeaSettings {\n\n"); 
		
		String base = "private static final String CLASS_NAME = ";
		String typeBase = "private static final String FILE_TYPE = ";
		if (DataType.getCurrentType() instanceof FileType ) {
			builder.append(base + "\"" + FileControl.class.getName() +"\";\n");	
			builder.append(typeBase + "\"file\";\n");	
		} else if (DataType.getCurrentType() instanceof ImageType ) {
			builder.append(base + "\"" + ImageControl.class.getName() +"\";\n");	
			builder.append(typeBase + "\"image\";\n");
		} else if (DataType.getCurrentType() instanceof EditorType ) {
			builder.append(base + "\"" + EditorControl.class.getName() +"\";\n");	
			builder.append(typeBase + "\"text file\";\n");
		} else if (DataType.getCurrentType() instanceof CalendarType ) {
			builder.append(base + "\"" + CalendarPEAControl.class.getName() +"\";\n");		
			builder.append(typeBase + "\"calendar\";\n");
		} 		
		
		builder.append("private static final boolean KEY_FILE_PROPERTY = " +  PeaFactory.isKeyFileProperty() + ";\n" );
		
		base = "private static final String EXTERNAL_FILE_PATH = ";
		builder.append(base + "null;\n" );

		if (JarStuff.getJarFileName() == null) {
			builder.append("private static final String JAR_FILE_NAME = \"default.jar\";\n");
		} else {
		builder.append("private static final String JAR_FILE_NAME = "
				+ "\"" + JarStuff.getJarFileName().substring(0, JarStuff.getJarFileName().length() -4) + "\";\n");
		}
		if ( (JarStuff.getLabelText() != null) 
				&& (JarStuff.getLabelText().length() > 0) ) { // if set: 
			if (JarStuff.getLabelText().length() > 40) {					
				JOptionPane.showMessageDialog(null,"Warning:\n" 					
						+ " Text above password-field longer than 40 characters will not be displayed. \n"
						+ " Encryption continues ..." );
			}
			builder.append("private static final String LABEL_TEXT = " + "\"" +  JarStuff.getLabelText() + "\";\n");

		} else {
			builder.append("private static final String LABEL_TEXT = null;\n" );
		}

		String programRandomString = "{";
		for (int i=0;i< Attachments.getProgramRandomBytesSize();i++) {
			programRandomString += Attachments.getProgramRandomBytes()[i] + ", ";
		}
		programRandomString = programRandomString.substring(0, programRandomString.length() - 2); // delete last ,
		builder.append("private static final byte[] PROGRAM_RANDOM_BYTES = " +  programRandomString + "};\n");
		//TestLog.o(FileModifier.class, "FILE ID: " + new String(Attachments.getFileIdentifier()));
		String fileIdentifierString = "{";
		for (int i=0;i<Attachments.getFileIdentifier().length;i++) {
			fileIdentifierString += Attachments.getFileIdentifier()[i] + ", ";
		}
		fileIdentifierString = fileIdentifierString.substring(0, fileIdentifierString.length() - 2); // delete last ,
		builder.append("private static final byte[] FILE_IDENTIFIER = " +  fileIdentifierString + "};\n");
		
		builder.append("private static final int PEA_VERSION = " +  PeaFactory.getPEA_VERSION() + ";\n");

		base = "private static final BlockCipher CIPHER_ALGO = ";
		BlockCipher cipherAlgo = CipherStuff.getCipherAlgo();
		if (cipherAlgo instanceof Shacal2Engine) {
			builder.append(base + "new Shacal2Engine();\n");
		} else if (cipherAlgo instanceof TwofishEngine) {
			builder.append(base + "new TwofishEngine();\n");
		} else if (cipherAlgo instanceof AESEngine) {
			builder.append(base + "new AESEngine();\n");
		} else if (cipherAlgo instanceof LEAEngine) {
			builder.append(base + "new LEAEngine();\n");
		} else if (cipherAlgo instanceof RC6Engine) {
			builder.append(base + "new RC6Engine();\n");
		} else if (cipherAlgo instanceof DSTU7624Engine) {
			builder.append(base + "new DSTU7624Engine(512);\n");
		} else if (cipherAlgo instanceof SerpentEngine) {
			builder.append(base + "new SerpentEngine();\n");
		} else if (cipherAlgo instanceof ThreefishEngine) {
			int blockSize = cipherAlgo.getBlockSize();
			builder.append(base + "new ThreefishEngine(" + blockSize * 8 + ");\n");
		} else {
			TestLog.e(FileModifier.class, "FileModifier: CipherAlgorithm invalid");
		}		
		// Hash algorithm
		base = "private static final Digest HASH_ALGO = ";
		Digest hashAlgo = HashStuff.getHashAlgo();
		if (hashAlgo instanceof WhirlpoolDigest) {
			builder.append(base + "new WhirlpoolDigest();\n");			
		} else if (hashAlgo instanceof SHA3Digest) {
			builder.append(base + "new SHA3Digest();\n");
		} else if (hashAlgo instanceof SkeinDigest) {
			builder.append(base + "new SkeinDigest(512, 512);\n");	
		} else if (hashAlgo instanceof SHA512Digest) {
			builder.append(base + "new SHA512Digest();\n");	
		} else if (hashAlgo instanceof Blake2bDigest) {
			builder.append(base + "new Blake2bDigest();\n");
		} else if (hashAlgo instanceof Blake3Digest) {
			builder.append(base + "new Blake3Digest();\n");
		} else {
			TestLog.e(FileModifier.class, "FileModifier: HashAlgorithm failed");
		}		
		// KDF scheme
		KeyDerivation kdfScheme = KeyDerivation.getKdf();
//System.out.println("FileModifier kdf: " + kdfScheme.getName());

		base = "private static final KeyDerivation KDF_SCHEME = ";
		if (kdfScheme instanceof BcryptKDF) {
			builder.append(base + "new BcryptKDF();\n");
		} else if (kdfScheme instanceof ScryptKDF) {
			builder.append(base + "new ScryptKDF();\n");
		} else if (kdfScheme instanceof PomeloKDF) {
			builder.append(base + "new PomeloKDF();\n");
		} else if (kdfScheme instanceof CatenaKDF) {
			builder.append(base + "new CatenaKDF();\n");
		} else if (kdfScheme instanceof Argon2KDF) {
			builder.append(base + "new Argon2KDF();\n");
		} else {
			TestLog.e(FileModifier.class, "FileModifier: Key derivation scheme invalid");
		}	

		builder.append("private static final int ITERATIONS = " +  KeyDerivation.getKdf().gettCost() + ";\n" );
		builder.append("private static final int MEMORY = " +  KeyDerivation.getKdf().getmCost() + ";\n" );
		builder.append("private static final int PARALLELIZATION = " +  KeyDerivation.getKdf().getArg3() + ";\n" );
		if (KeyDerivation.getVersionString().equals("")){
			builder.append("private static final String VERSION_STRING = \"\";\n" );
		} else {
			builder.append("private static final String VERSION_STRING = " + "\"" + KeyDerivation.getVersionString() + "\";\n" );
		}
		
		// Fixed content:
		builder.append(
				 "public final static Digest getHashAlgo() { return HASH_ALGO; }\n"
				+ "public final static int getIterations() { return ITERATIONS; }\n"
				+ "public final static int getMemory() { return MEMORY; }\n"
				+ "public final static int getParallelization() { return PARALLELIZATION; }\n"
				+ "public final static String getVersionString() { return VERSION_STRING; }\n"
				+ "public final static byte[] getProgramRandomBytes() { return PROGRAM_RANDOM_BYTES; }\n"
				+ "public final static byte[] getFileIdentifier() { return FILE_IDENTIFIER; }\n"	
				+ "public final static String getJarFileName() { return JAR_FILE_NAME; }\n"
				+ "public final static String getLabelText() { return LABEL_TEXT; }\n"
				+ "public final static boolean getKeyFileProperty() { return KEY_FILE_PROPERTY; }\n"
				+ "public final static String getExternalFilePath() { return EXTERNAL_FILE_PATH; }\n"
			//	+ "public final static void setExternalFilePath(String fileName) { EXTERNAL_FILE_PATH = fileName; }\n"				
				+ "public final static BlockCipher getCipherAlgo() { return CIPHER_ALGO; }\n"
				+ "public final static KeyDerivation getKdfScheme() { return KDF_SCHEME; }\n"
				+ "public final static int getPEA_VERSION() { return PEA_VERSION; }\n"
				+ "public final static String getCLASS_NAME() { return CLASS_NAME; }\n"
				+ "public final static String getFILE_TYPE() { return FILE_TYPE; }\n"
				
				+ "}\n"
				);
	    //TestLog.o(PeaSettings.class, new String(builder));
		BufferedWriter writer = null;

	    try {
			writer = new BufferedWriter( new FileWriter("settings" + File.separator + "PeaSettings.java"));
		    writer.write( new String(builder) );
		    writer.close();
		} catch (IOException e) {
			TestLog.e(FileModifier.class, e.getLocalizedMessage() + " - " + e.toString(), 5);
		}
	}

/*// FOR TESTING===========================
	public static void main(String[] args) {		
		//=====================
		// initialize (order!)
		//
	    if (HashStuff.getHashAlgo() == null) {
	    	HashStuff.setHashAlgo( new SkeinDigest(512, 512) );// default
	    }
		Attachments.generateAndSetProgramRandomBytes();
		Attachments.generateAndSetFileIdentifier();	
		KDFScheme.calculateSalt(Attachments.getProgramRandomBytes() );	    

	    if (CipherStuff.getCipherAlgo() == null) {
	    	CipherStuff.setCipherAlgo( new TwofishEngine() );
	    }
	    if (KDFScheme.getKDFScheme() == null) {
	    	KDFScheme.setKDFScheme( new SCryptConnector() );
	    }		
		FileModifier.generatePeaSettings();
		FileModifier.generateStart();		
		//new FileCompiler().compile(null);
	}*/
}
