package cologne.eck.peafactory.data;

/*
 * PeaFactory - Production of Password Encryption Archives
 * Copyright (C) 2022  Axel von dem Bruch
 * 
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, 
 * or (at your option) any later version.
 * This library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * See:  http://www.gnu.org/licenses/gpl-2.0.html
 * You should have received a copy of the GNU General Public License 
 * along with this library.
 */

import java.awt.Color;
import java.awt.Point;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.Window;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Properties;

import cologne.eck.peafactory.gui.MainView;
import cologne.eck.peafactory.gui.ProjectSelection;
import cologne.eck.tools.ExceptionHelper;
import cologne.eck.tools.TestLog;


public class FactoryPropertyHandler {


	// only file name, path inside jar is always /, not File.separator
	private static final String propFileName = "peafactory.properties";//"resources" + File.separator + 				

	/**
	 * Update all values in properties file.
	 * This method is called once when a PeaFactory is closed. 
	 * This should be called in EDT.
	 * 
	 * @return	error message or null
	 */
	public String updateAllPropValues() {

		String errorMessage = null;
		OutputStream outputStream = null;
		try {
			//==================================================
			Properties prop = new Properties();
			File file = new File(System.getProperty("user.dir") + File.separator 
					+ "resources" + File.separator + propFileName);
			if (! file.exists()) {
				// create directory
				file.getParentFile().mkdirs();
				// create file
				file.createNewFile();
			} 
			outputStream = new FileOutputStream(file);

			// Color ===============================
			Color peaColor = PeaFactoryProperties.getPeaColor();
			prop.setProperty("pea_color_r", "" + peaColor.getRed());
			prop.setProperty("pea_color_g", "" + peaColor.getGreen());
			prop.setProperty("pea_color_b", "" + peaColor.getBlue());

			// language =================================
			String lastLanguage = PeaFactoryProperties.getLastLanguage();
			if (lastLanguage != null) {
				prop.setProperty("language", lastLanguage);
			}

			// dimension and location ========================
			// get active window: MainView if open, ProjectSelection otherwise
			// if non of both is found, any open window
			Window[] windows = Window.getWindows();
			Window activeWin = null;
			boolean isMainView = false;
			for (Window w : windows) {
				if (w instanceof ProjectSelection && w.isShowing()) {
					activeWin = w;
				} else if (w instanceof MainView  && w.isShowing()) {
					activeWin = w;
					isMainView = true;
					break;
				}
			}
			if (activeWin == null) {
				for (Window w : windows) {
					if (w.isShowing()) {
						activeWin = w;
						break;
					}
				}
			}				
			if (activeWin == null) {
				// Fallback
				prop.setProperty("frame_x", "" + PeaFactoryProperties.getFrameLocation().getX());
				prop.setProperty("frame_y", "" + PeaFactoryProperties.getFrameLocation().getY());
				prop.setProperty("frame_width", "" + PeaFactoryProperties.getFrameDim().getWidth());
				prop.setProperty("frame_height", "" + PeaFactoryProperties.getFrameDim().getHeight());
			} else if (isMainView == true) {
				prop.setProperty("frame_x", "" + activeWin.getLocationOnScreen().getX());
				prop.setProperty("frame_y", "" + activeWin.getLocationOnScreen().getY());
				prop.setProperty("frame_width", "" + activeWin.getSize().getWidth());
				prop.setProperty("frame_height", "" + activeWin.getSize().getHeight());
			} else { // only location
				prop.setProperty("frame_x", "" + activeWin.getLocationOnScreen().getX());
				prop.setProperty("frame_y", "" + activeWin.getLocationOnScreen().getY());
				prop.setProperty("frame_width", "" + PeaFactoryProperties.getFrameDim().getWidth());
				prop.setProperty("frame_height", "" + PeaFactoryProperties.getFrameDim().getHeight());
			}
			prop.store(outputStream, "Properties for all PaeFactory");

			try {
				if (outputStream != null) {
					outputStream.close();
				}
			} catch (IOException e) {
				TestLog.ve(FactoryPropertyHandler.class, "FactoryPropertyHandler IOException: " + e.getMessage());
				errorMessage += e + ", " + e.getLocalizedMessage();;
			}

		} catch (IOException e) {
			TestLog.ve(FactoryPropertyHandler.class, "Exception when storig property values: " 
					+ new ExceptionHelper(e).getInfos());
			errorMessage += e + ", " + e.getLocalizedMessage();
		} catch (Exception e) {
			TestLog.ve(FactoryPropertyHandler.class, "Exception when storig property values: " 
					+ new ExceptionHelper(e).getInfos());
			errorMessage += e + ", " + e.getLocalizedMessage();			
		} finally {
			try {
				if (outputStream != null) {
					outputStream.close();
				}
			} catch (IOException e) {
				TestLog.ve(FactoryPropertyHandler.class, "FactoryPropertyHandler IOException: " + e.getMessage());
				errorMessage += e + ", " + e.getLocalizedMessage();;
			}
		}
		return errorMessage;
	}


	/**
	 * Read properties from the property file and set the
	 * values in PeaFactoryProperties. 
	 * This method is called once when a PeaFactory was started. 
	 * 
	 * @return	an error message or null
	 */
	public String readAndSetAllPropValues() {// throws IOException {
		String errorMessage = null;
		InputStream inputStream = null;
		try {
			//==================================================
			// set properties for all PEAs
			Properties prop = new Properties();

			File file = new File(System.getProperty("user.dir") + File.separator + "resources" + File.separator + propFileName);
			if (file.exists()) {
				// external file in resources
				inputStream = new FileInputStream(file);
			} else {
				// inside jar: default
				System.out.println("Load default properties...");
				inputStream = getClass().getResourceAsStream( "/resources/" +  propFileName);
			} 
			if (inputStream != null) {
				prop.load(inputStream);
			} else {// should not happen... 
				//throw new FileNotFoundException("property file '" + propFileName + "' not found in the classpath");
				TestLog.ve(FactoryPropertyHandler.class, "file not found: " + "/resources/" + propFileName);				
				errorMessage = "file not found: " + propFileName;		
				return errorMessage;
			}

			// should not happen:
			if (prop.getProperty("pea_color_r") == null
					&& prop.getProperty("font_size") == null
					&& prop.getProperty("frame_x") == null) {
				System.err.println("Can't read properties from file " + "resources" + File.separator + propFileName);
				TestLog.ve(FactoryPropertyHandler.class, "Color is missing", 5);
				return errorMessage + ("Can't read properties from file " + "resources" + File.separator + propFileName);
			}
			// Color ===============================
			int r, g, b;
			String rString = prop.getProperty("pea_color_r");
			try {
				r = Integer.parseInt(rString);
			} catch (Exception e) {
				r = PeaFactoryProperties.getPeaColor().getRed();
				TestLog.ve(FactoryPropertyHandler.class, "Color red is not digit: " + rString);
			}
			String gString = prop.getProperty("pea_color_g");
			try {
				g = Integer.parseInt(gString);
			} catch (Exception e) {
				g = PeaFactoryProperties.getPeaColor().getGreen();
				TestLog.ve(FactoryPropertyHandler.class, "Color green is not digit: " + gString);
			}
			String bString = prop.getProperty("pea_color_b");
			try {
				b = Integer.parseInt(bString);
			} catch (Exception e) {
				b = PeaFactoryProperties.getPeaColor().getBlue();
				TestLog.ve(FactoryPropertyHandler.class, "Color blue is not digit: " + bString);
			}
			Color color = new Color(r,g,b);
			PeaFactoryProperties.setPeaColor(color);

			// dimension and location ========================
			// avoid showing outside the monitor in Windows
			double screenWidth = 0;
			double screenHeight = 0;
			Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize(); 
			screenWidth = screenSize.getWidth();
			screenHeight = screenSize.getHeight();

			double w,h, x, y;			
			String frameX = prop.getProperty("frame_x");
			try {
				double xD = Double.parseDouble(frameX); 
				if (xD > screenWidth) {
					TestLog.ve(FactoryPropertyHandler.class, "Location x is outside the screen: " + xD + ", reset to default");
					xD = 50.0;
				}
				x = (int) xD;
			} catch (Exception e) {
				TestLog.ve(FactoryPropertyHandler.class, "Location x is not digit: " + frameX);
				x = PeaFactoryProperties.getFrameLocation().getX();
			}			
			String frameY = prop.getProperty("frame_y");
			try {
				double yD = Double.parseDouble(frameY);
				if (yD > screenHeight) {
					TestLog.ve(FactoryPropertyHandler.class, "Location y is outside the screen: " + yD + ", reset to default");
					yD = 50.0;
				}
				y = (int) yD;
			} catch (Exception e) {
				TestLog.ve(FactoryPropertyHandler.class, "Location yis not digit: " + frameY);
				y = PeaFactoryProperties.getFrameLocation().getY();
			}		
			String frameWidth = prop.getProperty("frame_width");
			if (frameWidth !=  null) {
				try {
					w = (int) Double.parseDouble(frameWidth);
				} catch (Exception e) {
					TestLog.ve(FactoryPropertyHandler.class, "Dimension width is not digit: " + frameWidth);
					w = PeaFactoryProperties.getFrameDim().getWidth();
				}
			} else {
				w = PeaFactoryProperties.getFrameDim().getWidth();
			}
			String frameHeight = prop.getProperty("frame_height");
			if (frameHeight != null) {
				try {
					h = (int) Double.parseDouble(frameHeight);
				} catch (Exception e) {
					TestLog.ve(FactoryPropertyHandler.class, "Dimension height is not digit: " + frameHeight);
					h = PeaFactoryProperties.getFrameDim().getHeight();
				}			
			} else {
				h = PeaFactoryProperties.getFrameDim().getHeight();
			}

			PeaFactoryProperties.setFrameDim(new Dimension((int) w, (int) h));
			PeaFactoryProperties.setFrameLocation(new Point( (int) x, (int) y));			


			String lastLanguage = prop.getProperty("language");
			if (lastLanguage != null) {
				PeaFactoryProperties.setLastLanguage(lastLanguage);
			}


		} catch (IOException e) {
			TestLog.ve(FactoryPropertyHandler.class, "FactoryPropertyHandler IOException: " + e.getMessage());
			errorMessage += e + ", " + e.getLocalizedMessage();
		} catch (Exception e) {
			TestLog.ve(FactoryPropertyHandler.class, "FactoryPropertyHandler Exception: " + e.getMessage());
			errorMessage += e + ", " + e.getLocalizedMessage();			
		} finally {
			try {
				if (inputStream != null) {
					inputStream.close();
				}
			} catch (IOException e) {
				TestLog.ve(FactoryPropertyHandler.class, "FactoryPropertyHandler IOException: " + e.getMessage());
				errorMessage += e + ", " + e.getLocalizedMessage();;
			}
		}
		return errorMessage;
	}
}