package cologne.eck.tools;

/*
 * PeaFactory - Production of Password Encryption Archives
 * Copyright (C) 2015  Axel von dem Bruch
 * 
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, 
 * or (at your option) any later version.
 * This library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * See:  http://www.gnu.org/licenses/gpl-2.0.html
 * You should have received a copy of the GNU General Public License 
 * along with this library.
 */

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;

import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.rendering.ImageType;
import org.apache.pdfbox.rendering.PDFRenderer;

import cologne.eck.peafactory.gui.MainView;

//import cologne.eck.peafactory.gui.MainView;


public class PDFConverter {
	

	/**
	 * higher values refer to better quality and
	 * more space
	 */
	private float scale = 4.0f; // quality
	
	/**
	 * The PDF file to convert
	 */
	private File pdfFile = null;
	
	/**
	 * the PDF file name with canonical path to convert
	 */
	private String pdfFileNameWithPath;
	
	/**
	 * the name of the file without path and without extension .pdf
	 */
	private String fileName;

	/**
	 * The file name of the PDF file without extension
	 * and with canonical path. This is used to create 
	 * the folder for the images. 
	 */
	private String destinationFolder;
	
	/**
	 * The document of the PDF file
	 */
	private PDDocument doc = null;
	
	/**
	 *  the number of all pages of the document
	 */
	private int allPages = 0;
	
	/**
	 * The format of the images to create:
	 *  JPEG, PNG, GIF, BMP. Default is PNG
	 */
	private String imageFormatName = "PNG";


	/**
	 * Instance of PDFConverter. Each instance loads
	 * a PDF file as its document. 
	 * 
	 * @param _pdfFile	the PDF file to convert
	 */
	public PDFConverter(File _pdfFile) {
		this.pdfFile = _pdfFile;
		try {
			this.pdfFileNameWithPath = _pdfFile.getCanonicalPath();
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		this.fileName = pdfFileNameWithPath.substring(
				pdfFileNameWithPath.lastIndexOf(File.separator), 
				pdfFileNameWithPath.length() - 4); // without extension .pdf
		this.destinationFolder = pdfFileNameWithPath.substring(
				0, 
				pdfFileNameWithPath.length() - 4);
		try {
			doc = PDDocument.load(pdfFile);
			allPages = doc.getNumberOfPages();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	/**
	 * Convert all pages of a PDF document to images. 
	 * The produced images are stored in a folder along with
	 * the PDF (named like the PDF file name (without extension .pdf). 
	 * Each page of the document is converted to one image. 
	 */
	public void convertAll() {

		convert(0, allPages);
	}
	
	/**
	 * Convert all pages of a PDF document to images. 
	 * The produced images are stored in a folder along with
	 * the PDF (named like the PDF file name (without extension .pdf). 
	 * Each page of the document is converted to one image. 
	 * 
	 * @param firstPage		the first page of the document to convert
	 * @param lastPage		the last page of the document to convert
	 * 
	 * @return				true if conversion was successful
	 */
	private boolean convert(int firstPage, int lastPage) {
				
		/*
		TODO
		Exceptions -> ErrorMessages
		*/

		int pageNumber = lastPage - firstPage;
		
		//progress of one page
		int progressOfOnePage = (100 - 10) / pageNumber;
		int progress = 10;

		PDFRenderer renderer = new PDFRenderer(doc);

		BufferedImage image = null;

		for (int i = 0; i < pageNumber; i++) {			
			
			try {
				//image = renderer.renderImageWithDPI(i, 100, ImageType.RGB);
				// set quality/scale of the image
				image = renderer.renderImage(i, scale, ImageType.RGB);
				
				if (new File(destinationFolder).exists()== false) {
					boolean success = new File(destinationFolder).mkdir();
					if (success == false) {
						System.err.println("Creation of folder for pdf conversion failed");
						return false;
					}
				}
				
			    // Java 7: ImageIOUtil.writeImage(image, pdfFileNameWithPath + "-" + (i) + "." + imageFormatName.toLowerCase(), 300);
				ImageIO.write(image, imageFormatName, new File(
						destinationFolder+ File.separator + fileName
						//pdfFileNameWithPath.substring(0,  pdfFile.getPath().length() -4) 
						+ "_" + (i + 1) + "."
						+ imageFormatName.toLowerCase()));
				
				//PDFConverterProgress.setValue(progress);
				MainView.progressBar.setValue(progress);
				MainView.progressBar.paint(MainView.progressBar.getGraphics());
				progress += progressOfOnePage;
				
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}			
		}
		try {
			doc.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		//PDFConverterProgress.setValue(0);
		MainView.progressBar.setValue(0);
		MainView.progressBar.paint(MainView.progressBar.getGraphics());

		return true;
	}

	/**
	 * Get the quality of the image. Larger values refer to
	 * better quality. Default value is 4.
	 * 
	 * @return the scale
	 */
	public float getScale() {
		return scale;
	}

	/**
	 * Set the quality of the image. Larger values refer to
	 * better quality. Default value is 4.
	 * 
	 * @param scalex the scale to set
	 */
	public void setScale(float newScalex) {
		this.scale = newScalex;
	}

	/**
	 * Get the used image format. 
	 * Default format is PNG. Other possible formats are
	 * JPEG, GIF, BMP and WBMP. 
	 * 
	 * @return the imageFormatName
	 */
	public String getImageFormatName() {
		return imageFormatName;
	}

	/**
	 * Set the image format to use. 
	 * Default format is PNG. Other possible formats are
	 * JPEG, GIF, BMP. If the format name is
	 * invalid, the previous name remains. 
	 * 
	 * @param imageFormatName the imageFormatName to set
	 */
	public void setImageFormatName(String newImageFormatName) {
		// JPEG, PNG, GIF, BMP and WBMP
		if (	newImageFormatName.equalsIgnoreCase("PNG") 
			|| 	newImageFormatName.equalsIgnoreCase("JPEG")
			|| 	newImageFormatName.equalsIgnoreCase("GIF")
			|| 	newImageFormatName.equalsIgnoreCase("BMP")){
			this.imageFormatName = newImageFormatName;
		} else {
			System.err.println("Invalid image format: " + newImageFormatName);
		}
	}
}
